"""Metrics analyzed using resampling methods."""

from __future__ import annotations

from collections.abc import Callable, Sequence
import functools
from typing import TYPE_CHECKING, NamedTuple

import numpy as np
import scipy.stats

import tea_tasting.config
from tea_tasting.metrics.base import MetricBaseGranular
import tea_tasting.utils


if TYPE_CHECKING:
    from typing import Literal

    import numpy.typing as npt
    import pyarrow as pa


class BootstrapResult(NamedTuple):
    """Result of the analysis using bootstrap resampling.

    Attributes:
        control: Control statistic value.
        treatment: Treatment statistic value.
        effect_size: Absolute effect size. Difference between the two statistic values.
        effect_size_ci_lower: Lower bound of the absolute effect size
            confidence interval.
        effect_size_ci_upper: Upper bound of the absolute effect size
            confidence interval.
        rel_effect_size: Relative effect size. Difference between the two statistic
            values, divided by the control statistic value.
        rel_effect_size_ci_lower: Lower bound of the relative effect size
            confidence interval.
        rel_effect_size_ci_upper: Upper bound of the relative effect size
            confidence interval.
    """
    control: float
    treatment: float
    effect_size: float
    effect_size_ci_lower: float
    effect_size_ci_upper: float
    rel_effect_size: float
    rel_effect_size_ci_lower: float
    rel_effect_size_ci_upper: float


class Bootstrap(MetricBaseGranular[BootstrapResult]):  # noqa: D101
    def __init__(
        self,
        columns: str | Sequence[str],
        statistic: Callable[..., npt.NDArray[np.number]],
        *,
        alternative: Literal["two-sided", "greater", "less"] | None = None,
        confidence_level: float | None = None,
        n_resamples: int | None = None,
        method: Literal["percentile", "basic", "bca"] = "bca",
        batch: int | None = None,
        random_state: int | np.random.Generator | np.random.SeedSequence | None = None,
    ) -> None:
        """Metric for analysis of a statistic using bootstrap resampling.

        If `columns` is a sequence of strings, then the sample passed
        to the statistic callable contains an extra dimension in the first axis.
        See examples below.

        Args:
            columns: Names of the columns to be used in the analysis.
            statistic: Statistic. It must be a vectorized callable
                that accepts a NumPy array as the first argument and returns
                the resulting statistic.
                It must also accept a keyword argument `axis` and be vectorized
                to compute the statistic along the provided axis.
            alternative: Alternative hypothesis:

                - `"two-sided"`: the means are unequal,
                - `"greater"`: the mean in the treatment variant is greater than the mean
                    in the control variant,
                - `"less"`: the mean in the treatment variant is less than the mean
                    in the control variant.

            confidence_level: Confidence level for the confidence interval.
            n_resamples: The number of resamples performed to form
                the bootstrap distribution of the statistic.
            method: Whether to return the "percentile" bootstrap confidence
                interval (`"percentile"`), the "basic" (AKA "reverse") bootstrap
                confidence interval (`"basic"`), or the bias-corrected
                and accelerated bootstrap confidence interval (`"bca"`).
            batch: The number of resamples to process in each vectorized call
                to statistic. Memory usage is O(`batch * n`), where `n` is
                the sample size. Default is `None`, in which case `batch = n_resamples`
                (or `batch = max(n_resamples, n)` for method="bca").
            random_state: Pseudorandom number generator state used
                to generate resamples.

        Parameter defaults:
            Defaults for parameters `alternative`, `confidence_level`,
            and `n_resamples` can be changed using the
            `config_context` and `set_context` functions.
            See the [Global configuration](https://tea-tasting.e10v.me/api/config/)
            reference for details.

        References:
            - [Bootstrapping (statistics) &#8212; Wikipedia](https://en.wikipedia.org/wiki/Bootstrapping_(statistics)).
            - [scipy.stats.bootstrap &#8212; SciPy Manual](https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.bootstrap.html#scipy-stats-bootstrap).

        Examples:
            ```pycon
            >>> import numpy as np
            >>> import tea_tasting as tt

            >>> experiment = tt.Experiment(
            ...     orders_per_user=tt.Bootstrap("orders", np.mean, random_state=42),
            ... )
            >>> data = tt.make_users_data(seed=42)
            >>> result = experiment.analyze(data)
            >>> result
                     metric control treatment rel_effect_size rel_effect_size_ci pvalue
            orders_per_user   0.530     0.573            8.0%       [-1.8%, 19%]      -

            ```

            With multiple columns:

            ```pycon
            >>> def ratio_of_means(sample, axis):
            ...     means = np.mean(sample, axis=axis)
            ...     return means[0] / means[1]

            >>> experiment = tt.Experiment(
            ...     orders_per_session=tt.Bootstrap(
            ...         ("orders", "sessions"),
            ...         ratio_of_means,
            ...         random_state=42,
            ...     ),
            ... )
            >>> data = tt.make_users_data(seed=42)
            >>> result = experiment.analyze(data)
            >>> result
                        metric control treatment rel_effect_size rel_effect_size_ci pvalue
            orders_per_session   0.266     0.289            8.8%      [-0.61%, 20%]      -

            ```
        """  # noqa: E501
        tea_tasting.utils.check_scalar(columns, "columns", typ=str | Sequence)
        if not isinstance(columns, str):
            for col in columns:
                tea_tasting.utils.check_scalar(col, "column", typ=str)
        self.columns = columns

        self.statistic = tea_tasting.utils.check_scalar(
            statistic, "statistic", typ=Callable)

        self.alternative = (
            tea_tasting.utils.auto_check(alternative, "alternative")
            if alternative is not None
            else tea_tasting.config.get_config("alternative")
        )

        self.confidence_level = (
            tea_tasting.utils.auto_check(confidence_level, "confidence_level")
            if confidence_level is not None
            else tea_tasting.config.get_config("confidence_level")
        )

        self.n_resamples = (
            tea_tasting.utils.auto_check(n_resamples, "n_resamples")
            if n_resamples is not None
            else tea_tasting.config.get_config("n_resamples")
        )

        self.method = tea_tasting.utils.check_scalar(
            method, "method", typ=str, in_={"percentile", "basic", "bca"})

        self.batch = tea_tasting.utils.check_scalar(batch, "batch", typ=int | None)

        self.random_state = tea_tasting.utils.check_scalar(
            random_state,
            "random_state",
            typ=int | np.random.Generator | np.random.SeedSequence | None,
        )


    @property
    def cols(self) -> Sequence[str]:
        """Columns to be fetched for a metric analysis."""
        if isinstance(self.columns, str):
            return (self.columns,)
        return self.columns


    def analyze_granular(
        self,
        control: pa.Table,
        treatment: pa.Table,
    ) -> BootstrapResult:
        """Analyze metric in an experiment using granular data.

        Args:
            control: Control data.
            treatment: Treatment data.

        Returns:
            Analysis result.
        """
        def statistic(
            contr: npt.NDArray[np.number],
            treat: npt.NDArray[np.number],
            axis: int = -1,
        ) -> npt.NDArray[np.number]:
            contr_stat = self.statistic(contr, axis=axis)
            treat_stat = self.statistic(treat, axis=axis)

            effect_size = treat_stat - contr_stat
            with np.errstate(divide="ignore", invalid="ignore"):
                rel_effect_size = np.divide(treat_stat, contr_stat) - 1

            return np.stack((effect_size, rel_effect_size), axis=0)

        contr = _select_as_numpy(control, self.columns)
        treat = _select_as_numpy(treatment, self.columns)
        stat = statistic(contr, treat, axis=0)

        result = scipy.stats.bootstrap(
            (contr, treat),
            statistic,
            n_resamples=self.n_resamples,
            batch=self.batch,
            axis=0,
            confidence_level=self.confidence_level,
            alternative=self.alternative,
            method=self.method,
            random_state=self.random_state,  # type: ignore
        )
        ci = result.confidence_interval

        return BootstrapResult(
            control=self.statistic(contr, axis=0),  # type: ignore
            treatment=self.statistic(treat, axis=0),  # type: ignore
            effect_size=stat[0],
            effect_size_ci_lower=ci.low[0],
            effect_size_ci_upper=ci.high[0],
            rel_effect_size=stat[1],
            rel_effect_size_ci_lower=ci.low[1],
            rel_effect_size_ci_upper=ci.high[1],
        )


def _select_as_numpy(
    data: pa.Table,
    columns: str | Sequence[str],
) -> npt.NDArray[np.number]:
    if isinstance(columns, str):
        return data[columns].combine_chunks().to_numpy(zero_copy_only=False)
    return np.column_stack([
        data[col].combine_chunks().to_numpy(zero_copy_only=False)
        for col in columns
    ])


class Quantile(Bootstrap):  # noqa: D101
    def __init__(
        self,
        column: str,
        q: float = 0.5,
        *,
        alternative: Literal["two-sided", "greater", "less"] | None = None,
        confidence_level: float | None = None,
        n_resamples: int | None = None,
        method: Literal["percentile", "basic", "bca"] = "basic",
        batch: int | None = None,
        random_state: int | np.random.Generator | np.random.SeedSequence | None = None,
    ) -> None:
        """Metric for the analysis of quantiles using bootstrap resampling.

        Args:
            column: Name of the column for the quantiles to compute.
            q: Probability for the quantiles to compute.
            alternative: Alternative hypothesis:

                - `"two-sided"`: the means are unequal,
                - `"greater"`: the mean in the treatment variant is greater than the mean
                    in the control variant,
                - `"less"`: the mean in the treatment variant is less than the mean
                    in the control variant.

            confidence_level: Confidence level for the confidence interval.
            n_resamples: The number of resamples performed to form
                the bootstrap distribution of the statistic.
            method: Whether to return the "percentile" bootstrap confidence
                interval (`"percentile"`), the "basic" (AKA "reverse") bootstrap
                confidence interval (`"basic"`), or the bias-corrected
                and accelerated bootstrap confidence interval (`"bca"`).

                Default method is "basic" which is different from default
                method "bca" in `Bootstrap`. The "bca" confidence intervals cannot
                be calculated when the bootstrap distribution is degenerate
                (e.g. all elements are identical). This is often the case for the
                quantile metrics.

            batch: The number of resamples to process in each vectorized call
                to statistic. Memory usage is O(`batch * n`), where `n` is
                the sample size. Default is `None`, in which case `batch = n_resamples`
                (or `batch = max(n_resamples, n)` for method="bca").
            random_state: Pseudorandom number generator state used
                to generate resamples.

        Parameter defaults:
            Defaults for parameters `alternative`, `confidence_level`,
            and `n_resamples` can be changed using the
            `config_context` and `set_context` functions.
            See the [Global configuration](https://tea-tasting.e10v.me/api/config/)
            reference for details.

        Examples:
            ```pycon
            >>> import tea_tasting as tt

            >>> experiment = tt.Experiment(
            ...     revenue_per_user_p80=tt.Quantile("revenue", 0.8, random_state=42),
            ... )
            >>> data = tt.make_users_data(seed=42)
            >>> result = experiment.analyze(data)
            >>> result
                          metric control treatment rel_effect_size rel_effect_size_ci pvalue
            revenue_per_user_p80    10.6      11.6            9.1%       [-1.2%, 21%]      -

            ```
        """  # noqa: E501
        self.column = tea_tasting.utils.check_scalar(column, "column", typ=str)
        self.q = tea_tasting.utils.check_scalar(q, "q", typ=float, ge=0, le=1)
        super().__init__(
            columns=column,
            statistic=functools.partial(np.nanquantile, q=q),
            alternative=alternative,
            confidence_level=confidence_level,
            n_resamples=n_resamples,
            method=method,
            batch=batch,
            random_state=random_state,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, NamedTuple
import unittest.mock

import ibis
import polars as pl
import pyarrow as pa
import pyarrow.compute as pc
import pytest

import tea_tasting.aggr
import tea_tasting.datasets
import tea_tasting.metrics.base


if TYPE_CHECKING:
    from typing import Any, Literal

    import ibis.expr.types  # noqa: TC004
    import pandas as pd


    Frame = ibis.expr.types.Table | pa.Table | pd.DataFrame | pl.LazyFrame


def test_aggr_cols_or():
    aggr_cols0 = tea_tasting.metrics.base.AggrCols(
        has_count=False,
        mean_cols=("a", "b"),
        var_cols=("b", "c"),
        cov_cols=(("a", "b"), ("c", "b")),
    )

    aggr_cols1 = tea_tasting.metrics.base.AggrCols(
        has_count=True,
        mean_cols=("b", "c"),
        var_cols=("c", "d"),
        cov_cols=(("b", "c"), ("d", "c")),
    )

    aggr_cols = aggr_cols0 | aggr_cols1

    assert isinstance(aggr_cols, tea_tasting.metrics.base.AggrCols)
    assert aggr_cols.has_count is True
    assert set(aggr_cols.mean_cols) == {"a", "b", "c"}
    assert len(aggr_cols.mean_cols) == 3
    assert set(aggr_cols.var_cols) == {"b", "c", "d"}
    assert len(aggr_cols.var_cols) == 3
    assert set(aggr_cols.cov_cols) == {("a", "b"), ("b", "c"), ("c", "d")}
    assert len(aggr_cols.cov_cols) == 3


def test_aggr_cols_len():
    assert len(tea_tasting.metrics.base.AggrCols(
        has_count=False,
        mean_cols=("a", "b"),
        var_cols=("b", "c"),
        cov_cols=(("a", "b"), ("c", "b")),
    )) == 6
    assert len(tea_tasting.metrics.base.AggrCols(
        has_count=True,
        mean_cols=("b", "c"),
        var_cols=("c", "d"),
        cov_cols=(("b", "c"), ("d", "c")),
    )) == 7


@pytest.fixture
def data_arrow() -> pa.Table:
    return tea_tasting.datasets.make_users_data(n_users=100, seed=42)

@pytest.fixture
def data_pandas(data_arrow: pa.Table) -> pd.DataFrame:
    return data_arrow.to_pandas()

@pytest.fixture
def data_polars(data_arrow: pa.Table) -> pl.DataFrame:
    return pl.from_arrow(data_arrow)  # type: ignore

@pytest.fixture
def data_polars_lazy(data_polars: pl.DataFrame) -> pl.LazyFrame:
    return data_polars.lazy()

@pytest.fixture
def data_duckdb(data_arrow: pa.Table) -> ibis.expr.types.Table:
    return ibis.connect("duckdb://").create_table("data", data_arrow)

@pytest.fixture
def data_sqlite(data_arrow: pa.Table) -> ibis.expr.types.Table:
    return ibis.connect("sqlite://").create_table("data", data_arrow)

@pytest.fixture(params=[
    "data_arrow", "data_pandas",
    "data_polars", "data_polars_lazy",
    "data_duckdb", "data_sqlite",
])
def data(request: pytest.FixtureRequest) -> Frame:
    return request.getfixturevalue(request.param)


@pytest.fixture
def aggr_cols() -> tea_tasting.metrics.base.AggrCols:
    return tea_tasting.metrics.base.AggrCols(
        has_count=True,
        mean_cols=("sessions", "orders"),
        var_cols=("orders", "revenue"),
        cov_cols=(("sessions", "revenue"),),
    )

@pytest.fixture
def correct_aggrs(
    data_arrow: pa.Table,
    aggr_cols: tea_tasting.metrics.base.AggrCols,
) -> dict[object, tea_tasting.aggr.Aggregates]:
    return tea_tasting.aggr.read_aggregates(
        data_arrow,
        group_col="variant",
        **aggr_cols._asdict(),
    )

@pytest.fixture
def correct_aggr(
    data_arrow: pa.Table,
    aggr_cols: tea_tasting.metrics.base.AggrCols,
) -> tea_tasting.aggr.Aggregates:
    return tea_tasting.aggr.read_aggregates(
        data_arrow,
        group_col=None,
        **aggr_cols._asdict(),
    )

@pytest.fixture
def cols() -> tuple[str, ...]:
    return ("sessions", "orders", "revenue")

@pytest.fixture
def correct_gran(
    data_arrow: pa.Table,
    cols: tuple[str, ...],
) -> dict[object, pa.Table]:
    variant_col = data_arrow["variant"]
    table = data_arrow.select(cols)
    return {
        var: table.filter(pc.equal(variant_col, pa.scalar(var)))  # type: ignore
        for var in variant_col.unique().to_pylist()
    }

@pytest.fixture
def aggr_metric(
    aggr_cols: tea_tasting.metrics.base.AggrCols,
) -> tea_tasting.metrics.base.MetricBaseAggregated[dict[str, object]]:
    class AggrMetric(tea_tasting.metrics.base.MetricBaseAggregated[dict[str, object]]):
        @property
        def aggr_cols(self) -> tea_tasting.metrics.base.AggrCols:
            return aggr_cols

        def analyze_aggregates(
            self,
            control: tea_tasting.aggr.Aggregates,  # noqa: ARG002
            treatment: tea_tasting.aggr.Aggregates,  # noqa: ARG002
        ) -> dict[str, object]:
            return {}

    return AggrMetric()

@pytest.fixture
def aggr_power(
    aggr_cols: tea_tasting.metrics.base.AggrCols,
) -> tea_tasting.metrics.base.PowerBaseAggregated[
    tea_tasting.metrics.base.MetricPowerResults[dict[str, object]]
]:
    class AggrPower(
        tea_tasting.metrics.base.PowerBaseAggregated[
            tea_tasting.metrics.base.MetricPowerResults[dict[str, object]]
        ],
    ):
        @property
        def aggr_cols(self) -> tea_tasting.metrics.base.AggrCols:
            return aggr_cols

        def solve_power_from_aggregates(
            self,
            data: tea_tasting.aggr.Aggregates,  # noqa: ARG002
            parameter: Literal[  # noqa: ARG002
                "power",
                "effect_size",
                "rel_effect_size",
                "n_obs",
            ] = "power",
        ) -> tea_tasting.metrics.base.MetricPowerResults[dict[str, object]]:
            return tea_tasting.metrics.base.MetricPowerResults()
    return AggrPower()

@pytest.fixture
def gran_metric(
    cols: tuple[str, ...],
) -> tea_tasting.metrics.base.MetricBaseGranular[dict[str, object]]:
    class GranMetric(tea_tasting.metrics.base.MetricBaseGranular[dict[str, object]]):
        @property
        def cols(self) -> tuple[str, ...]:
            return cols

        def analyze_granular(
            self,
            control: pa.Table,  # noqa: ARG002
            treatment: pa.Table,  # noqa: ARG002
        ) -> dict[str, object]:
            return {}

    return GranMetric()


def _compare_aggrs(
    left: tea_tasting.aggr.Aggregates,
    right: tea_tasting.aggr.Aggregates,
) -> None:
    assert left.count_ == right.count_
    assert left.mean_ == pytest.approx(right.mean_)
    assert left.var_ == pytest.approx(right.var_)
    assert left.cov_ == pytest.approx(right.cov_)


def test_metric_power_results_to_dicts():
    result0 = {
        "power": 0.8,
        "effect_size": 1,
        "rel_effect_size": 0.05,
        "n_obs": 10_000,
    }
    result1 = {
        "power": 0.9,
        "effect_size": 2,
        "rel_effect_size": 0.1,
        "n_obs": 20_000,
    }

    results = tea_tasting.metrics.base.MetricPowerResults[dict[str, float | int]](  # type: ignore
        [result0, result1])
    assert results.to_dicts() == (result0, result1)

    class PowerResult(NamedTuple):
        power: float
        effect_size: float
        rel_effect_size: float
        n_obs: float
    results = tea_tasting.metrics.base.MetricPowerResults[PowerResult]([
        PowerResult(**result0),
        PowerResult(**result1),
    ])
    assert results.to_dicts() == (result0, result1)


def test_metric_base_aggregated_analyze_frame(
    aggr_metric: tea_tasting.metrics.base.MetricBaseAggregated[dict[str, object]],
    data_arrow: pa.Table,
    correct_aggrs: dict[object, tea_tasting.aggr.Aggregates],
):
    aggr_metric.analyze_aggregates = unittest.mock.MagicMock()
    aggr_metric.analyze(data_arrow, control=0, treatment=1, variant="variant")
    aggr_metric.analyze_aggregates.assert_called_once()
    kwargs = aggr_metric.analyze_aggregates.call_args.kwargs
    _compare_aggrs(kwargs["control"], correct_aggrs[0])
    _compare_aggrs(kwargs["treatment"], correct_aggrs[1])

def test_metric_base_aggregated_analyze_aggrs(
    aggr_metric: tea_tasting.metrics.base.MetricBaseAggregated[dict[str, object]],
    correct_aggrs: dict[object, tea_tasting.aggr.Aggregates],
):
    aggr_metric.analyze_aggregates = unittest.mock.MagicMock()
    aggr_metric.analyze(correct_aggrs, control=0, treatment=1)
    aggr_metric.analyze_aggregates.assert_called_once()
    kwargs = aggr_metric.analyze_aggregates.call_args.kwargs
    _compare_aggrs(kwargs["control"], correct_aggrs[0])
    _compare_aggrs(kwargs["treatment"], correct_aggrs[1])


def test_power_base_aggregated_analyze_frame(
    aggr_power: tea_tasting.metrics.base.PowerBaseAggregated[Any],
    data_arrow: pa.Table,
    correct_aggr: tea_tasting.aggr.Aggregates,
):
    aggr_power.solve_power_from_aggregates = unittest.mock.MagicMock()
    aggr_power.solve_power(data_arrow, "effect_size")
    aggr_power.solve_power_from_aggregates.assert_called_once()
    kwargs = aggr_power.solve_power_from_aggregates.call_args.kwargs
    _compare_aggrs(kwargs["data"], correct_aggr)
    assert kwargs["parameter"] == "effect_size"

def test_power_base_aggregated_analyze_aggr(
    aggr_power: tea_tasting.metrics.base.PowerBaseAggregated[Any],
    correct_aggr: tea_tasting.aggr.Aggregates,
):
    aggr_power.solve_power_from_aggregates = unittest.mock.MagicMock()
    aggr_power.solve_power(correct_aggr, "rel_effect_size")
    aggr_power.solve_power_from_aggregates.assert_called_once()
    kwargs = aggr_power.solve_power_from_aggregates.call_args.kwargs
    _compare_aggrs(kwargs["data"], correct_aggr)
    assert kwargs["parameter"] == "rel_effect_size"


def test_aggregate_by_variants_frame(
    data_arrow: pa.Table,
    aggr_cols: tea_tasting.metrics.base.AggrCols,
    correct_aggrs: dict[object, tea_tasting.aggr.Aggregates],
):
    aggrs = tea_tasting.metrics.base.aggregate_by_variants(
        data_arrow,
        aggr_cols=aggr_cols,
        variant="variant",
    )
    _compare_aggrs(aggrs[0], correct_aggrs[0])
    _compare_aggrs(aggrs[1], correct_aggrs[1])

def test_aggregate_by_variants_aggrs(
    aggr_cols: tea_tasting.metrics.base.AggrCols,
    correct_aggrs: dict[object, tea_tasting.aggr.Aggregates],
):
    aggrs = tea_tasting.metrics.base.aggregate_by_variants(
        correct_aggrs,
        aggr_cols=aggr_cols,
        variant="variant",
    )
    _compare_aggrs(aggrs[0], correct_aggrs[0])
    _compare_aggrs(aggrs[1], correct_aggrs[1])

def test_aggregate_by_variants_raises(
    data_arrow: pa.Table,
    aggr_cols: tea_tasting.metrics.base.AggrCols,
):
    with pytest.raises(ValueError, match="variant"):
        tea_tasting.metrics.base.aggregate_by_variants(data_arrow, aggr_cols=aggr_cols)


def test_metric_base_granular_frame(
    gran_metric: tea_tasting.metrics.base.MetricBaseGranular[dict[str, object]],
    data_arrow: pa.Table,
    correct_gran: dict[object, pa.Table],
):
    gran_metric.analyze_granular = unittest.mock.MagicMock()
    gran_metric.analyze(data_arrow, control=0, treatment=1, variant="variant")
    gran_metric.analyze_granular.assert_called_once()
    kwargs = gran_metric.analyze_granular.call_args.kwargs
    assert kwargs["control"].equals(correct_gran[0])
    assert kwargs["treatment"].equals(correct_gran[1])

def test_metric_base_granular_gran(
    gran_metric: tea_tasting.metrics.base.MetricBaseGranular[dict[str, object]],
    correct_gran: dict[object, pa.Table],
):
    gran_metric.analyze_granular = unittest.mock.MagicMock()
    gran_metric.analyze(correct_gran, control=0, treatment=1)
    gran_metric.analyze_granular.assert_called_once()
    kwargs = gran_metric.analyze_granular.call_args.kwargs
    assert kwargs["control"].equals(correct_gran[0])
    assert kwargs["treatment"].equals(correct_gran[1])


def test_read_granular_frame(
    data: Frame,
    cols: tuple[str, ...],
    correct_gran: dict[object, pa.Table],
):
    gran = tea_tasting.metrics.base.read_granular(
        data,
        cols=cols,
        variant="variant",
    )
    assert gran[0].equals(correct_gran[0])
    assert gran[1].equals(correct_gran[1])

def test_read_granular_dict(
    cols: tuple[str, ...],
    correct_gran: dict[object, pa.Table],
):
    gran = tea_tasting.metrics.base.read_granular(
        correct_gran,
        cols=cols,
        variant="variant",
    )
    assert gran[0].equals(correct_gran[0])
    assert gran[1].equals(correct_gran[1])

def test_read_granular_none(
    data: Frame,
    cols: tuple[str, ...],
    data_arrow: pa.Table,
):
    gran = tea_tasting.metrics.base.read_granular(data, cols=cols)
    assert gran.equals(data_arrow.select(cols))



"""A Python package for the statistical analysis of A/B tests.

All classes and functions for the analysis of the experiments can be imported
from the root `tea_tasting` module.

There are functions and classes for advanced use cases such as defining custom metrics.
They can be imported from submodules of `tea_tasting`.

For convenience, the API reference is provided by submodules:

- `tea_tasting.metrics`: Built-in metrics.
- `tea_tasting.experiment`: Experiment and experiment result.
- `tea_tasting.multiplicity`: Multiple hypothesis testing.
- `tea_tasting.datasets`: Example datasets.
- `tea_tasting.config`: Global configuration.
- `tea_tasting.aggr`: Module for working with aggregated statistics.
- `tea_tasting.utils`: Useful functions and classes.
"""
# pyright: reportUnusedImport=false

from tea_tasting.config import config_context, get_config, set_config
from tea_tasting.datasets import make_sessions_data, make_users_data
from tea_tasting.experiment import Experiment
from tea_tasting.metrics import Bootstrap, Mean, Quantile, RatioOfMeans, SampleRatio
from tea_tasting.multiplicity import adjust_fdr, adjust_fwer
from tea_tasting.version import __version__



# pyright: reportAttributeAccessIssue=false
from __future__ import annotations

import concurrent.futures
from typing import TYPE_CHECKING, NamedTuple, TypedDict

import ibis
import ibis.expr.types
import polars as pl
import pyarrow as pa
import pyarrow.compute as pc
import pytest

import tea_tasting.aggr
import tea_tasting.datasets
import tea_tasting.experiment
import tea_tasting.metrics
import tea_tasting.utils


if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from typing import Literal

    import narwhals.typing
    import numpy as np
    import pandas as pd


    Frame = ibis.expr.types.Table | pa.Table | pd.DataFrame | pl.LazyFrame


class _MetricResultTuple(NamedTuple):
    control: float
    treatment: float
    effect_size: float

class _MetricResultDict(TypedDict):
    control: float
    treatment: float
    effect_size: float

class _PowerResult(NamedTuple):
    power: float
    effect_size: float
    rel_effect_size: float
    n_obs: float


class _Metric(
    tea_tasting.metrics.MetricBase[_MetricResultTuple],
    tea_tasting.metrics.PowerBase[tea_tasting.metrics.MetricPowerResults[_PowerResult]],
):
    def __init__(self, value: str) -> None:
        self.value = value

    def analyze(
        self,
        data: narwhals.typing.IntoFrame | ibis.expr.types.Table | dict[
            object, tea_tasting.aggr.Aggregates],
        control: object,
        treatment: object,
        variant: str,
    ) -> _MetricResultTuple:
        if not isinstance(data, dict):
            data = tea_tasting.aggr.read_aggregates(
                data,
                variant,
                has_count=False,
                mean_cols=(self.value,),
                var_cols=(),
                cov_cols=(),
            )
        return _MetricResultTuple(
            control=data[control].mean(self.value),
            treatment=data[treatment].mean(self.value),
            effect_size=data[treatment].mean(self.value) -
                data[control].mean(self.value),
        )

    def solve_power(
        self,
        data: narwhals.typing.IntoFrame | ibis.Table,  # noqa: ARG002
        parameter: Literal[  # noqa: ARG002
            "power", "effect_size", "rel_effect_size", "n_obs"] = "rel_effect_size",
    ) -> tea_tasting.metrics.MetricPowerResults[_PowerResult]:
        return tea_tasting.metrics.MetricPowerResults((
            _PowerResult(power=0.8, effect_size=1, rel_effect_size=0.05, n_obs=10_000),
            _PowerResult(power=0.9, effect_size=2, rel_effect_size=0.1, n_obs=20_000),
        ))


class _MetricAggregated(
    tea_tasting.metrics.MetricBaseAggregated[_MetricResultTuple],
    tea_tasting.metrics.PowerBaseAggregated[
        tea_tasting.metrics.MetricPowerResults[dict[str, object]]],
):
    def __init__(self, value: str) -> None:
        self.value = value

    @property
    def aggr_cols(self) -> tea_tasting.metrics.AggrCols:
        return tea_tasting.metrics.AggrCols(mean_cols=(self.value,))

    def analyze_aggregates(
        self,
        control: tea_tasting.aggr.Aggregates,
        treatment: tea_tasting.aggr.Aggregates,
    ) -> _MetricResultTuple:
        contr_mean = control.mean(self.value)
        treat_mean = treatment.mean(self.value)
        return _MetricResultTuple(
            control=contr_mean,
            treatment=treat_mean,
            effect_size=treat_mean - contr_mean,
        )

    def solve_power_from_aggregates(
        self,
        data: tea_tasting.aggr.Aggregates,  # noqa: ARG002
        parameter: Literal[  # noqa: ARG002
            "power", "effect_size", "rel_effect_size", "n_obs"] = "rel_effect_size",
    ) -> tea_tasting.metrics.MetricPowerResults[dict[str, object]]:
        return tea_tasting.metrics.MetricPowerResults((
            {"power": 0.8, "effect_size": 1, "rel_effect_size": 0.05, "n_obs": 10_000},
            {"power": 0.9, "effect_size": 2, "rel_effect_size": 0.1, "n_obs": 20_000},
        ))


class _MetricGranular(tea_tasting.metrics.MetricBaseGranular[_MetricResultDict]):  # type: ignore
    def __init__(self, value: str) -> None:
        self.value = value

    @property
    def cols(self) -> tuple[str, ...]:
        return (self.value,)

    def analyze_granular(
        self,
        control: pa.Table,
        treatment: pa.Table,
    ) -> _MetricResultDict:
        contr_mean = pc.mean(control[self.value]).as_py()  # type: ignore
        treat_mean = pc.mean(treatment[self.value]).as_py()  # type: ignore
        return _MetricResultDict(
            control=contr_mean,
            treatment=treat_mean,
            effect_size=treat_mean - contr_mean,
        )


@pytest.fixture
def get_result() -> Callable[
    [tuple[float, float, float], tuple[float, float, float]],
    tea_tasting.experiment.ExperimentResult,
]:
    def _get_result(
        x: tuple[float, float, float],
        y: tuple[float, float, float],
    ) -> tea_tasting.experiment.ExperimentResult:
        return tea_tasting.experiment.ExperimentResult(
            metric_tuple=_MetricResultTuple(*x),
            metric_dict=_MetricResultDict(
                control=y[0],
                treatment=y[1],
                effect_size=y[2],
            ),  # type: ignore
        )

    return _get_result


@pytest.fixture
def result(
    get_result: Callable[
        [tuple[float, float, float], tuple[float, float, float]],
        tea_tasting.experiment.ExperimentResult,
    ],
) -> tea_tasting.experiment.ExperimentResult:
    return get_result((10, 11, 1), (20, 22, 2))


@pytest.fixture
def result2() -> tea_tasting.experiment.ExperimentResult:
    class MetricResultTuple(NamedTuple):
        control: float
        treatment: float
        rel_effect_size: float
        rel_effect_size_ci_lower: float
        rel_effect_size_ci_upper: float
        pvalue: float

    return tea_tasting.experiment.ExperimentResult(
        metric_tuple=MetricResultTuple(
            control=4.4444,
            treatment=5.5555,
            rel_effect_size=0.2,
            rel_effect_size_ci_lower=0.12345,
            rel_effect_size_ci_upper=float("inf"),
            pvalue=0.23456,
        ),
        metric_dict={
            "control": 9.9999,
            "treatment": 11.111,
            "rel_effect_size": 0.11111,
            "rel_effect_size_ci_lower": 0,
        },
    )


@pytest.fixture
def results(
    result: tea_tasting.experiment.ExperimentResult,
) -> tea_tasting.experiment.ExperimentResults:
    return tea_tasting.experiment.ExperimentResults({(0, 1): result})


@pytest.fixture
def results2(
    get_result: Callable[
        [tuple[float, float, float], tuple[float, float, float]],
        tea_tasting.experiment.ExperimentResult,
    ],
) -> tea_tasting.experiment.ExperimentResults:
    return tea_tasting.experiment.ExperimentResults({
        (0, 1): get_result((10, 11, 1), (20, 22, 2)),
        (0, 2): get_result((10, 11, 1), (30, 33, 3)),
    })


@pytest.fixture
def data_arrow() -> pa.Table:
    return tea_tasting.datasets.make_users_data(n_users=100, seed=42)

@pytest.fixture
def data_pandas(data_arrow: pa.Table) -> pd.DataFrame:
    return data_arrow.to_pandas()

@pytest.fixture
def data_polars(data_arrow: pa.Table) -> pl.DataFrame:
    return pl.from_arrow(data_arrow)  # type: ignore

@pytest.fixture
def data_polars_lazy(data_polars: pl.DataFrame) -> pl.LazyFrame:
    return data_polars.lazy()

@pytest.fixture
def data_duckdb(data_arrow: pa.Table) -> ibis.expr.types.Table:
    return ibis.connect("duckdb://").create_table("data", data_arrow)

@pytest.fixture
def data_sqlite(data_arrow: pa.Table) -> ibis.expr.types.Table:
    return ibis.connect("sqlite://").create_table("data", data_arrow)

@pytest.fixture(params=[
    "data_arrow", "data_pandas",
    "data_polars", "data_polars_lazy",
    "data_duckdb", "data_sqlite",
])
def data(request: pytest.FixtureRequest) -> Frame:
    return request.getfixturevalue(request.param)

@pytest.fixture
def data_arrow_multi(data_arrow: pa.Table) -> pa.Table:
    data2 = data_arrow.filter(pc.equal(data_arrow["variant"], pa.scalar(1)))
    return pa.concat_tables((
        data_arrow,
        data2.set_column(
            data_arrow.schema.get_field_index("variant"),
            "variant",
            pa.array([2] * data2.num_rows),
        ),
    ))


@pytest.fixture
def ref_result(
    data_arrow: pa.Table,
) -> tea_tasting.experiment.ExperimentResults:
    sessions = _Metric("sessions")
    orders = _MetricAggregated("orders")
    revenue = _MetricGranular("revenue")
    return tea_tasting.experiment.ExperimentResult(
        avg_sessions=sessions.analyze(data_arrow, 0, 1, "variant"),
        avg_orders=orders.analyze(data_arrow, 0, 1, "variant"),
        avg_revenue=revenue.analyze(data_arrow, 0, 1, "variant"),  # type: ignore
    )


def test_experiment_result_to_dicts(result: tea_tasting.experiment.ExperimentResult):
    assert result.to_dicts() == (
        {"metric": "metric_tuple", "control": 10, "treatment": 11, "effect_size": 1},
        {"metric": "metric_dict", "control": 20, "treatment": 22, "effect_size": 2},
    )


def test_experiment_results_to_dicts(
    results2: tea_tasting.experiment.ExperimentResults,
):
    assert results2.to_dicts() == (
        {
            "variants": "(0, 1)",
            "metric": "metric_tuple",
            "control": 10,
            "treatment": 11,
            "effect_size": 1,
        },
        {
            "variants": "(0, 1)",
            "metric": "metric_dict",
            "control": 20,
            "treatment": 22,
            "effect_size": 2,
        },
        {
            "variants": "(0, 2)",
            "metric": "metric_tuple",
            "control": 10,
            "treatment": 11,
            "effect_size": 1,
        },
        {
            "variants": "(0, 2)",
            "metric": "metric_dict",
            "control": 30,
            "treatment": 33,
            "effect_size": 3,
        },
    )


def test_simulation_results_to_dicts(
    results2: tea_tasting.experiment.ExperimentResults,
):
    assert tea_tasting.experiment.SimulationResults(results2.values()).to_dicts() == (
        {
            "metric": "metric_tuple",
            "control": 10,
            "treatment": 11,
            "effect_size": 1,
        },
        {
            "metric": "metric_dict",
            "control": 20,
            "treatment": 22,
            "effect_size": 2,
        },
        {
            "metric": "metric_tuple",
            "control": 10,
            "treatment": 11,
            "effect_size": 1,
        },
        {
            "metric": "metric_dict",
            "control": 30,
            "treatment": 33,
            "effect_size": 3,
        },
    )


def test_experiment_power_result_to_dicts():
    raw_results = (
        {"power": 0.8, "effect_size": 1, "rel_effect_size": 0.05, "n_obs": 20_000},
        {"power": 0.9, "effect_size": 1, "rel_effect_size": 0.05, "n_obs": 10_000},
        {"power": 0.8, "effect_size": 2, "rel_effect_size": 0.1, "n_obs": 10_000},
        {"power": 0.9, "effect_size": 2, "rel_effect_size": 0.1, "n_obs": 20_000},
    )
    result = tea_tasting.experiment.ExperimentPowerResult({
        "metric_dict": tea_tasting.metrics.MetricPowerResults[dict[str, float | int]](  # type: ignore
            raw_results[0:2]),
        "metric_tuple": tea_tasting.metrics.MetricPowerResults[_PowerResult]([
            _PowerResult(**raw_results[2]),
            _PowerResult(**raw_results[3]),
        ]),
    })
    assert isinstance(result, tea_tasting.utils.DictsReprMixin)
    assert result.default_keys == (
        "metric", "power", "effect_size", "rel_effect_size", "n_obs")
    assert result.to_dicts() == (
        {"metric": "metric_dict", "power": 0.8, "effect_size": 1,
            "rel_effect_size": 0.05, "n_obs": 20_000},
        {"metric": "metric_dict", "power": 0.9, "effect_size": 1,
            "rel_effect_size": 0.05, "n_obs": 10_000},
        {"metric": "metric_tuple", "power": 0.8, "effect_size": 2,
            "rel_effect_size": 0.1, "n_obs": 10_000},
        {"metric": "metric_tuple", "power": 0.9, "effect_size": 2,
            "rel_effect_size": 0.1, "n_obs": 20_000},
    )


def test_experiment_init_default():
    metrics = {
        "avg_sessions": _Metric("sessions"),
        "avg_orders": _MetricAggregated("orders"),
        "avg_revenue": _MetricGranular("revenue"),
    }
    experiment = tea_tasting.experiment.Experiment(metrics)  # type: ignore
    assert experiment.metrics == metrics
    assert experiment.variant == "variant"

def test_experiment_init_kwargs():
    metrics = {
        "avg_sessions": _Metric("sessions"),
        "avg_orders": _MetricAggregated("orders"),
        "avg_revenue": _MetricGranular("revenue"),
    }
    experiment = tea_tasting.experiment.Experiment(**metrics)  # type: ignore
    assert experiment.metrics == metrics
    assert experiment.variant == "variant"

def test_experiment_init_custom():
    metrics = {
        "avg_sessions": _Metric("sessions"),
        "avg_orders": _MetricAggregated("orders"),
        "avg_revenue": _MetricGranular("revenue"),
    }
    experiment = tea_tasting.experiment.Experiment(metrics, "group")  # type: ignore
    assert experiment.metrics == metrics
    assert experiment.variant == "group"


def test_experiment_analyze_default(
    data: Frame,
    ref_result: tea_tasting.experiment.ExperimentResult,
):
    experiment = tea_tasting.experiment.Experiment({
        "avg_sessions": _Metric("sessions"),
        "avg_orders": _MetricAggregated("orders"),
        "avg_revenue": _MetricGranular("revenue"),
    })
    assert experiment.analyze(data) == ref_result

def test_experiment_analyze_base(
    data: Frame,
    ref_result: tea_tasting.experiment.ExperimentResult,
):
    experiment = tea_tasting.experiment.Experiment({
        "avg_sessions": _Metric("sessions"),
    })
    assert experiment.analyze(data) == tea_tasting.experiment.ExperimentResult(
        avg_sessions=ref_result["avg_sessions"])

def test_experiment_analyze_aggr(
    data: Frame,
    ref_result: tea_tasting.experiment.ExperimentResult,
):
    experiment = tea_tasting.experiment.Experiment({
        "avg_orders": _MetricAggregated("orders"),
    })
    assert experiment.analyze(data) == tea_tasting.experiment.ExperimentResult(
        avg_orders=ref_result["avg_orders"])

def test_experiment_analyze_gran(
    data: Frame,
    ref_result: tea_tasting.experiment.ExperimentResult,
):
    experiment = tea_tasting.experiment.Experiment({
        "avg_revenue": _MetricGranular("revenue"),
    })
    assert experiment.analyze(data) == tea_tasting.experiment.ExperimentResult(
        avg_revenue=ref_result["avg_revenue"])

def test_experiment_analyze_all_pairs(
    data_arrow_multi: pa.Table,
    ref_result: tea_tasting.experiment.ExperimentResult,
):
    experiment = tea_tasting.experiment.Experiment({
        "avg_sessions": _Metric("sessions"),
        "avg_orders": _MetricAggregated("orders"),
        "avg_revenue": _MetricGranular("revenue"),
    })
    results = experiment.analyze(data_arrow_multi, all_variants=True)
    assert set(results.keys()) == {(0, 1), (0, 2), (1, 2)}
    assert results[0, 1] == ref_result
    assert results[0, 2] == ref_result

def test_experiment_analyze_all_pairs_raises(data_arrow_multi: pa.Table):
    experiment = tea_tasting.experiment.Experiment({
        "avg_sessions": _Metric("sessions"),
        "avg_orders": _MetricAggregated("orders"),
        "avg_revenue": _MetricGranular("revenue"),
    })
    with pytest.raises(ValueError, match="all_variants"):
        experiment.analyze(data_arrow_multi)

def test_experiment_analyze_two_treatments(
    data_arrow_multi: pa.Table,
    ref_result: tea_tasting.experiment.ExperimentResult,
):
    experiment = tea_tasting.experiment.Experiment(
        {
            "avg_sessions": _Metric("sessions"),
            "avg_orders": _MetricAggregated("orders"),
            "avg_revenue": _MetricGranular("revenue"),
        },
    )
    results = experiment.analyze(data_arrow_multi, control=0, all_variants=True)
    assert results == tea_tasting.experiment.ExperimentResults({
        (0, 1): ref_result,
        (0, 2): ref_result,
    })


def test_experiment_solve_power(data_arrow: pa.Table):
    experiment = tea_tasting.experiment.Experiment(
        metric=_Metric("sessions"),
        metric_aggr=_MetricAggregated("orders"),
    )
    result = experiment.solve_power(data_arrow)
    assert result == tea_tasting.experiment.ExperimentPowerResult({
        "metric": tea_tasting.metrics.MetricPowerResults((
            _PowerResult(power=0.8, effect_size=1, rel_effect_size=0.05, n_obs=10_000),
            _PowerResult(power=0.9, effect_size=2, rel_effect_size=0.1, n_obs=20_000),
        )),
        "metric_aggr": tea_tasting.metrics.MetricPowerResults((
            {"power": 0.8, "effect_size": 1, "rel_effect_size": 0.05, "n_obs": 10_000},
            {"power": 0.9, "effect_size": 2, "rel_effect_size": 0.1, "n_obs": 20_000},
        )),
    })


class ExperimentWithSimulationResults(tea_tasting.experiment.Experiment):
    def __init__(self, *args: object, **kwargs: object) -> None:
        self.simulation_results = tea_tasting.experiment.SimulationResults()
        self.data = []
        super().__init__(*args, **kwargs)  # type: ignore

    def analyze(  # type: ignore
        self,
        *args: object,
        **kwargs: object,
    ) -> tea_tasting.experiment.ExperimentResult:
        self.data.append(args[0])
        result = super().analyze(*args, **kwargs)  # type: ignore
        self.simulation_results.append(result)
        return result

def test_experiment_simulate_default(data: Frame):
    experiment = ExperimentWithSimulationResults({
        "avg_sessions": _MetricAggregated("sessions"),
        "avg_orders": _MetricGranular("orders"),
        "avg_revenue": _Metric("revenue"),
    })
    assert experiment.simulate(data, 10, seed=42) == experiment.simulation_results

def test_experiment_simulate_cols(data: Frame, data_arrow: pa.Arrow):
    experiment = ExperimentWithSimulationResults({
        "avg_sessions": _MetricAggregated("sessions"),
    })
    assert experiment.simulate(data, 10, seed=42) == experiment.simulation_results
    assert experiment.data[0].column_names == ["sessions", "variant"]
    experiment = ExperimentWithSimulationResults({
        "avg_orders": _MetricGranular("orders"),
    })
    assert experiment.simulate(data, 10, seed=42) == experiment.simulation_results
    assert experiment.data[0].column_names == ["orders", "variant"]
    experiment = ExperimentWithSimulationResults({
        "avg_revenue": _Metric("revenue"),
    })
    assert experiment.simulate(data, 10, seed=42) == experiment.simulation_results
    assert set(experiment.data[0].column_names) == set(data_arrow.column_names)

def test_experiment_simulate_callable():
    experiment = ExperimentWithSimulationResults({
        "avg_sessions": _MetricAggregated("sessions"),
        "avg_orders": _MetricGranular("orders"),
        "avg_revenue": _Metric("revenue"),
    })
    tables = []
    def make_data(seed: np.random.Generator) -> pa.Table:
        table = tea_tasting.datasets.make_users_data(seed=seed, n_users=100)
        tables.append(table)
        return table
    results = experiment.simulate(make_data, 10, seed=42)
    assert results == experiment.simulation_results
    assert results[0] == experiment.analyze(tables[0])

def test_experiment_simulate_map(data_arrow: pa.Table):
    experiment = tea_tasting.experiment.Experiment({
        "avg_sessions": _MetricAggregated("sessions"),
        "avg_orders": _MetricGranular("orders"),
        "avg_revenue": _Metric("revenue"),
    })
    with concurrent.futures.ProcessPoolExecutor() as executor:
        results = experiment.simulate(data_arrow, 10, seed=42, map_=executor.map)
    assert results == experiment.simulate(data_arrow, 10, seed=42)

def test_experiment_simulate_progress(data_arrow: pa.Table):
    experiment = tea_tasting.experiment.Experiment({
        "avg_sessions": _MetricAggregated("sessions"),
        "avg_orders": _MetricGranular("orders"),
        "avg_revenue": _Metric("revenue"),
    })
    simulation_results = tea_tasting.experiment.SimulationResults()
    def progress(
        results: Iterable[tea_tasting.experiment.ExperimentResult],
    ) -> Iterable[tea_tasting.experiment.ExperimentResult]:
        results = tuple(results)
        simulation_results.extend(results)
        return results
    results = experiment.simulate(data_arrow, 10, seed=42, progress=progress)
    assert results == simulation_results

def test_experiment_simulate_treat():
    experiment = tea_tasting.experiment.Experiment({
        "avg_sessions": _MetricAggregated("sessions"),
        "avg_orders": _MetricGranular("orders"),
        "avg_revenue": _Metric("revenue"),
    })
    data = tea_tasting.datasets.make_users_data(
        seed=42,
        n_users=1000,
        orders_uplift=0,
        revenue_uplift=0,
    )
    def treat(data: pa.Table) -> pa.Table:
        return (
            data.drop_columns(["orders", "revenue"])
            .append_column("orders", pc.multiply(data["orders"], pa.scalar(1.1)))  # type: ignore
            .append_column("revenue", pc.multiply(data["revenue"], pa.scalar(1.1)))  # type: ignore
        )
    results = experiment.simulate(data, 100, seed=42, treat=treat)
    means = (
        results.to_polars()
        .select("metric", "effect_size")
        .group_by("metric").mean()
        .sort("metric")
    )
    assert means.item(2, "effect_size") == pytest.approx(0, abs=0.001)
    assert means.item(0, "effect_size") == pytest.approx(0.05, rel=0.1)
    assert means.item(1, "effect_size") == pytest.approx(0.5, rel=0.1)



"""Experiment and experiment result."""

from __future__ import annotations

from collections import UserDict, UserList
import functools
import itertools
from typing import TYPE_CHECKING, Any, overload

import ibis.expr.types
import narwhals as nw
import numpy as np
import pyarrow as pa
import pyarrow.compute as pc

import tea_tasting.aggr
import tea_tasting.metrics
import tea_tasting.utils


if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from typing import Concatenate, Literal, TypeAlias, TypeVar

    import narwhals.typing  # noqa: TC004


    T = TypeVar("T")
    MapLike: TypeAlias = Callable[Concatenate[Callable[..., T], ...], Iterable[T]]
    ProgressFn: TypeAlias = Callable[Concatenate[Iterable[T], ...], Iterable[T]]
    DataGenerator: TypeAlias =  Callable[
        ..., narwhals.typing.IntoFrame | ibis.expr.types.Table]


class ExperimentResult(
    tea_tasting.utils.DictsReprMixin,
    UserDict[str, tea_tasting.metrics.MetricResult],
):
    """Experiment result for a pair of variants."""
    default_keys = (
        "metric",
        "control",
        "treatment",
        "rel_effect_size",
        "rel_effect_size_ci",
        "pvalue",
    )

    @tea_tasting.utils._cache_method
    def to_dicts(self) -> tuple[dict[str, object], ...]:
        """Convert the result to a sequence of dictionaries.

        Examples:
            ```pycon
            >>> import pprint
            >>> import tea_tasting as tt

            >>> experiment = tt.Experiment(
            ...     orders_per_user=tt.Mean("orders"),
            ...     revenue_per_user=tt.Mean("revenue"),
            ... )
            >>> data = tt.make_users_data(seed=42)
            >>> result = experiment.analyze(data)
            >>> pprint.pprint(result.to_dicts())
            ({'control': 0.5304003954522986,
              'effect_size': 0.04269014577177832,
              'effect_size_ci_lower': -0.010800201598205515,
              'effect_size_ci_upper': 0.09618049314176216,
              'metric': 'orders_per_user',
              'pvalue': np.float64(0.11773177998716214),
              'rel_effect_size': 0.08048664016431273,
              'rel_effect_size_ci_lower': -0.019515294044061937,
              'rel_effect_size_ci_upper': 0.1906880061278886,
              'statistic': 1.5647028839586707,
              'treatment': 0.5730905412240769},
             {'control': 5.241028175976273,
              'effect_size': 0.4890831037404775,
              'effect_size_ci_lower': -0.13261881482742033,
              'effect_size_ci_upper': 1.1107850223083753,
              'metric': 'revenue_per_user',
              'pvalue': np.float64(0.1230698855425058),
              'rel_effect_size': 0.09331815958981626,
              'rel_effect_size_ci_lower': -0.02373770894855798,
              'rel_effect_size_ci_upper': 0.22440926894909308,
              'statistic': 1.5423440700784083,
              'treatment': 5.73011127971675})

            ```
        """
        return tuple(
            {"metric": k} | (v if isinstance(v, dict) else v._asdict())
            for k, v in self.items()
        )  # type: ignore


class ExperimentResults(
    tea_tasting.utils.DictsReprMixin,
    UserDict[tuple[object, object], ExperimentResult],
):
    """Experiment results for multiple pairs of variants."""
    default_keys = (
        "variants",
        "metric",
        "control",
        "treatment",
        "rel_effect_size",
        "rel_effect_size_ci",
        "pvalue",
    )

    @tea_tasting.utils._cache_method
    def to_dicts(self) -> tuple[dict[str, object], ...]:
        """Convert the results to a sequence of dictionaries."""
        return tuple(
            {"variants": str(variants)} | metric_result
            for variants, experiment_result in self.items()
            for metric_result in experiment_result.to_dicts()
        )


class SimulationResults(tea_tasting.utils.DictsReprMixin, UserList[ExperimentResult]):
    """Simulation results.

    Simulations are not enumerated for better performance.
    """
    default_keys = (
        "metric",
        "control",
        "treatment",
        "rel_effect_size",
        "rel_effect_size_ci",
        "pvalue",
    )
    default_max_rows: int = 10

    @tea_tasting.utils._cache_method
    def to_dicts(self) -> tuple[dict[str, object], ...]:
        """Convert the results to a sequence of dictionaries."""
        return tuple(itertools.chain.from_iterable(
            experiment_result.to_dicts()
            for experiment_result in self
        ))


class ExperimentPowerResult(
    tea_tasting.utils.DictsReprMixin,
    UserDict[str, tea_tasting.metrics.MetricPowerResults[Any]],
):
    """Result of the analysis of power in a experiment."""
    default_keys = ("metric", "power", "effect_size", "rel_effect_size", "n_obs")

    @tea_tasting.utils._cache_method
    def to_dicts(self) -> tuple[dict[str, object], ...]:
        """Convert the result to a sequence of dictionaries."""
        dicts = ()
        for metric, results in self.items():
            dicts = (*dicts, *({"metric": metric} | d for d in results.to_dicts()))
        return dicts


class Experiment(tea_tasting.utils.ReprMixin):  # noqa: D101
    def __init__(
        self,
        metrics: dict[str, tea_tasting.metrics.MetricBase[Any]] | None = None,
        variant: str = "variant",
        **kw_metrics: tea_tasting.metrics.MetricBase[Any],
    ) -> None:
        """Experiment definition: metrics and variant column.

        Args:
            metrics: Dictionary of metrics with metric names as keys.
            variant: Variant column name.
            kw_metrics: Metrics with metric names as parameter names.

        Examples:
            ```pycon
            >>> import tea_tasting as tt

            >>> experiment = tt.Experiment(
            ...     sessions_per_user=tt.Mean("sessions"),
            ...     orders_per_session=tt.RatioOfMeans("orders", "sessions"),
            ...     orders_per_user=tt.Mean("orders"),
            ...     revenue_per_user=tt.Mean("revenue"),
            ... )
            >>> data = tt.make_users_data(seed=42)
            >>> result = experiment.analyze(data)
            >>> result
                        metric control treatment rel_effect_size rel_effect_size_ci pvalue
             sessions_per_user    2.00      1.98          -0.66%      [-3.7%, 2.5%]  0.674
            orders_per_session   0.266     0.289            8.8%      [-0.89%, 19%] 0.0762
               orders_per_user   0.530     0.573            8.0%       [-2.0%, 19%]  0.118
              revenue_per_user    5.24      5.73            9.3%       [-2.4%, 22%]  0.123

            ```

            Using the first argument `metrics` which accepts metrics in a form of dictionary:

            ```pycon
            >>> experiment = tt.Experiment({
            ...     "sessions per user": tt.Mean("sessions"),
            ...     "orders per session": tt.RatioOfMeans("orders", "sessions"),
            ...     "orders per user": tt.Mean("orders"),
            ...     "revenue per user": tt.Mean("revenue"),
            ... })
            >>> data = tt.make_users_data(seed=42)
            >>> result = experiment.analyze(data)
            >>> result
                        metric control treatment rel_effect_size rel_effect_size_ci pvalue
             sessions per user    2.00      1.98          -0.66%      [-3.7%, 2.5%]  0.674
            orders per session   0.266     0.289            8.8%      [-0.89%, 19%] 0.0762
               orders per user   0.530     0.573            8.0%       [-2.0%, 19%]  0.118
              revenue per user    5.24      5.73            9.3%       [-2.4%, 22%]  0.123

            ```

            Power analysis:

            ```pycon
            >>> data = tt.make_users_data(
            ...     seed=42,
            ...     sessions_uplift=0,
            ...     orders_uplift=0,
            ...     revenue_uplift=0,
            ...     covariates=True,
            ... )
            >>> with tt.config_context(n_obs=(10_000, 20_000)):
            ...     experiment = tt.Experiment(
            ...         sessions_per_user=tt.Mean("sessions", "sessions_covariate"),
            ...         orders_per_session=tt.RatioOfMeans(
            ...             numer="orders",
            ...             denom="sessions",
            ...             numer_covariate="orders_covariate",
            ...             denom_covariate="sessions_covariate",
            ...         ),
            ...         orders_per_user=tt.Mean("orders", "orders_covariate"),
            ...         revenue_per_user=tt.Mean("revenue", "revenue_covariate"),
            ...     )
            >>> power_result = experiment.solve_power(data)
            >>> power_result
                        metric power effect_size rel_effect_size n_obs
             sessions_per_user   80%      0.0458            2.3% 10000
             sessions_per_user   80%      0.0324            1.6% 20000
            orders_per_session   80%      0.0177            6.8% 10000
            orders_per_session   80%      0.0125            4.8% 20000
               orders_per_user   80%      0.0374            7.2% 10000
               orders_per_user   80%      0.0264            5.1% 20000
              revenue_per_user   80%       0.488            9.2% 10000
              revenue_per_user   80%       0.345            6.5% 20000

            ```
        """  # noqa: E501
        if metrics is None:
            metrics = {}
        metrics = metrics | kw_metrics

        tea_tasting.utils.check_scalar(metrics, "metrics", typ=dict)
        tea_tasting.utils.check_scalar(len(metrics), "len(metrics)", gt=0)
        for name, metric in metrics.items():
            tea_tasting.utils.check_scalar(name, "metric_name", typ=str)
            tea_tasting.utils.check_scalar(
                metric, name, typ=tea_tasting.metrics.MetricBase)

        self.metrics = metrics
        self.variant = tea_tasting.utils.check_scalar(
            variant, "variant", typ=str)


    @overload
    def analyze(
        self,
        data: narwhals.typing.IntoFrame | ibis.expr.types.Table,
        control: object = None,
        *,
        all_variants: Literal[False] = False,
    ) -> ExperimentResult:
        ...

    @overload
    def analyze(
        self,
        data: narwhals.typing.IntoFrame | ibis.expr.types.Table,
        control: object = None,
        *,
        all_variants: Literal[True],
    ) -> ExperimentResults:
        ...

    def analyze(
        self,
        data: narwhals.typing.IntoFrame | ibis.expr.types.Table,
        control: object = None,
        *,
        all_variants: bool = False,
    ) -> ExperimentResult | ExperimentResults:
        """Analyze the experiment.

        Args:
            data: Experimental data.
            control: Control variant. If `None`, the variant with the minimal ID
                is used as a control.
            all_variants: If `True`, analyze all pairs of variants. Otherwise,
                analyze only one pair of variants.

        Returns:
            Experiment result.
        """
        tea_tasting.utils.check_scalar(all_variants, "all_variants", typ=bool)
        aggregated_data, granular_data = self._read_data(data)

        if aggregated_data is not None:
            variants = aggregated_data.keys()
        elif granular_data is not None:
            variants = granular_data.keys()
        else:
            variants = self._read_variants(data)
        variants = sorted(variants)  # type: ignore

        if control is not None:
            variant_pairs = tuple(
                (control, treatment)
                for treatment in variants
                if treatment != control
            )
        else:
            variant_pairs = tuple(
                (control, treatment)
                for control in variants
                for treatment in variants
                if control < treatment
            )

        if len(variant_pairs) != 1 and not all_variants:
            raise ValueError(
                "all_variants is False, but there are more than one pair of variants.")

        results = ExperimentResults()
        for contr, treat in variant_pairs:
            result = ExperimentResult()
            for name, metric in self.metrics.items():
                result |= {name: self._analyze_metric(
                    metric=metric,
                    data=data,
                    aggregated_data=aggregated_data,
                    granular_data=granular_data,
                    control=contr,
                    treatment=treat,
                )}

            if not all_variants:
                return result

            results |= {(contr, treat): result}

        return results


    def _analyze_metric(
        self,
        metric: tea_tasting.metrics.MetricBase[Any],
        data: narwhals.typing.IntoFrame | ibis.expr.types.Table,
        aggregated_data: dict[object, tea_tasting.aggr.Aggregates] | None,
        granular_data: dict[object, pa.Table] | None,
        control: object,
        treatment: object,
    ) -> tea_tasting.metrics.MetricResult:
        if (
            isinstance(metric, tea_tasting.metrics.MetricBaseAggregated)
            and aggregated_data is not None
        ):
            return metric.analyze(aggregated_data, control, treatment)

        if (
            isinstance(metric, tea_tasting.metrics.MetricBaseGranular)
            and granular_data is not None
        ):
            return metric.analyze(granular_data, control, treatment)

        return metric.analyze(data, control, treatment, self.variant)


    def _read_data(
        self,
        data: narwhals.typing.IntoFrame | ibis.expr.types.Table,
    ) -> tuple[
        dict[object, tea_tasting.aggr.Aggregates] | None,
        dict[object, pa.Table] | None,
    ]:
        aggr_cols = tea_tasting.metrics.AggrCols()
        gran_cols = set()
        for metric in self.metrics.values():
            if isinstance(metric, tea_tasting.metrics.MetricBaseAggregated):
                aggr_cols |= metric.aggr_cols
            if isinstance(metric, tea_tasting.metrics.MetricBaseGranular):
                gran_cols |= set(metric.cols)

        aggregated_data = tea_tasting.metrics.aggregate_by_variants(
            data,
            aggr_cols=aggr_cols,
            variant=self.variant,
        ) if len(aggr_cols) > 0 else None

        granular_data = tea_tasting.metrics.read_granular(
            data,
            cols=tuple(gran_cols),
            variant=self.variant,
        ) if len(gran_cols) > 0 else None

        return aggregated_data, granular_data


    def _read_variants(
        self,
        data: narwhals.typing.IntoFrame | ibis.expr.types.Table,
    ) -> list[object]:
        if isinstance(data, ibis.expr.types.Table):
            return (
                data.select(self.variant)
                .distinct()
                .to_pyarrow()[self.variant]
                .to_pylist()
            )

        data = nw.from_native(data)
        if not isinstance(data, nw.LazyFrame):
            data = data.lazy()
        return data.unique(self.variant).collect().get_column(self.variant).to_list()


    def solve_power(
        self,
        data: narwhals.typing.IntoFrame | ibis.expr.types.Table,
        parameter: Literal[
            "power", "effect_size", "rel_effect_size", "n_obs"] = "rel_effect_size",
    ) -> ExperimentPowerResult:
        """Solve for a parameter of the power of a test.

        Args:
            data: Sample data.
            parameter: Parameter name.

        Returns:
            Power analysis result.
        """
        tea_tasting.utils.check_scalar(
            parameter,
            "parameter",
            in_={"power", "effect_size", "rel_effect_size", "n_obs"},
        )
        aggr_cols = tea_tasting.metrics.AggrCols()
        for metric in self.metrics.values():
            if isinstance(metric, tea_tasting.metrics.PowerBaseAggregated):
                aggr_cols |= metric.aggr_cols

        aggr_data = tea_tasting.aggr.read_aggregates(
            data,
            group_col=None,
            **aggr_cols._asdict(),
        ) if len(aggr_cols) > 0 else tea_tasting.aggr.Aggregates()

        result = ExperimentPowerResult()
        for name, metric in self.metrics.items():
            if isinstance(metric, tea_tasting.metrics.PowerBaseAggregated):
                result |= {name: metric.solve_power(aggr_data, parameter=parameter)}
            elif isinstance(metric, tea_tasting.metrics.PowerBase):
                result |= {name: metric.solve_power(data, parameter=parameter)}

        return result


    def simulate(
        self,
        data: narwhals.typing.IntoFrame | ibis.expr.types.Table | DataGenerator,  # type: ignore
        n_simulations: int = 10_000,
        *,
        seed: int | np.random.Generator | np.random.SeedSequence | None = None,
        ratio: float | int = 1,
        treat: Callable[[pa.Table], pa.Table] | None = None,
        map_: MapLike[Any] = map,
        progress: ProgressFn[Any] | type[Iterable[Any]] | None = None,
    ) -> SimulationResults:
        """Simulate the experiment analysis multiple times.

        Args:
            data: Experimental data or a callable that generates the data.
            n_simulations: Number of simulations.
            seed: Random seed.
            ratio: Ratio of the number of users in treatment relative to control.
            treat: Treatment function that takes a PyArrow Table as an input
                and returns an updated PyArrow Table.
            map_: Map-like function to run simulations.
            progress: tqdm-like class or function to show the progress of simulations.

        Returns:
            Simulation results.
        """
        tea_tasting.utils.check_scalar(n_simulations, "n_simulations", typ=int, gt=0)
        tea_tasting.utils.auto_check(ratio, "ratio")

        if not callable(data):
            gran_cols: set[str] = set()
            for metric in self.metrics.values():
                if isinstance(metric, tea_tasting.metrics.MetricBaseAggregated):
                    aggr_cols = metric.aggr_cols
                    gran_cols |= (
                        set(aggr_cols.mean_cols) |
                        set(aggr_cols.var_cols) |
                        set(itertools.chain.from_iterable(aggr_cols.cov_cols))
                    )
                elif isinstance(metric, tea_tasting.metrics.MetricBaseGranular):
                    gran_cols |= set(metric.cols)
                else:
                    gran_cols = set()
                    break
            cols = tuple(gran_cols)
            data: pa.Table = tea_tasting.metrics.read_granular(data, cols)
            if self.variant in data.column_names:
                data = data.drop_columns(self.variant)

        sim = functools.partial(
            _simulate_once,
            experiment=self,
            data=data,
            ratio=ratio,
            treat=treat,
        )

        results = map_(sim, np.random.default_rng(seed).spawn(n_simulations))
        if progress is not None:
            try:
                results = progress(results, total=n_simulations)  # type: ignore
            except TypeError:
                results = progress(results)  # type: ignore
        return SimulationResults(results)


def _simulate_once(
    rng: np.random.Generator,
    experiment: Experiment,
    data: pa.Table | DataGenerator,  # type: ignore
    ratio: float | int,
    treat: Callable[[pa.Table], pa.Table] | None,
) -> ExperimentResult:
    if callable(data):
        data: pa.Table = tea_tasting.metrics.read_granular(data(seed=rng))  # type: ignore

    if experiment.variant not in data.column_names:
        data = data.append_column(
            experiment.variant,
            [rng.binomial(n=1, p=ratio / (1 + ratio), size=data.num_rows)],
        )

    if treat is not None:
        variant_array = data[experiment.variant]
        contr_data = data.filter(pc.equal(variant_array, pa.scalar(0)))  # type: ignore
        treat_data = treat(data.filter(pc.equal(variant_array, pa.scalar(1))))  # type: ignore
        if not contr_data.schema.equals(treat_data.schema):
            schema = pa.unify_schemas(
                [contr_data.schema, treat_data.schema],
                promote_options="permissive",
            )
            contr_data = contr_data.select(schema.names).cast(schema)
            treat_data = treat_data.select(schema.names).cast(schema)
        data = pa.concat_tables((contr_data, treat_data))

    return experiment.analyze(data)



"""Multiple hypothesis testing."""

from __future__ import annotations

import abc
from collections import UserDict
from typing import TYPE_CHECKING

import tea_tasting.config
import tea_tasting.experiment
import tea_tasting.utils


if TYPE_CHECKING:
    from collections.abc import Callable, Mapping, Sequence
    from typing import Literal


NO_NAME_COMPARISON = "-"


class MultipleComparisonsResults(
    tea_tasting.utils.DictsReprMixin,
    UserDict[object, tea_tasting.experiment.ExperimentResult],
):
    """Multiple comparisons result."""
    default_keys = (
        "comparison",
        "metric",
        "control",
        "treatment",
        "rel_effect_size",
        "pvalue",
        "pvalue_adj",
    )

    @tea_tasting.utils._cache_method
    def to_dicts(self) -> tuple[dict[str, object], ...]:
        """Convert the result to a sequence of dictionaries."""
        return tuple(
            {"comparison": str(comparison)} | metric_result
            for comparison, experiment_result in self.items()
            for metric_result in experiment_result.to_dicts()
        )


def adjust_fdr(
    experiment_results: tea_tasting.experiment.ExperimentResult | Mapping[
        object, tea_tasting.experiment.ExperimentResult],
    metrics: str | set[str] | Sequence[str] | None = None,
    *,
    alpha: float | None = None,
    arbitrary_dependence: bool = False,
) -> MultipleComparisonsResults:
    """Adjust p-value and alpha to control the false discovery rate (FDR).

    The number of hypotheses tested is the total number of metrics included in
    the comparison in all experiment results. For example, if there are
    3 experiments with 2 metrics in each, the number of hypotheses is 6.

    The function performs one of the following corrections, depending on parameters:

    - Benjamini-Hochberg procedure, assuming non-negative correlation between
        hypotheses (`arbitrary_dependence=False`).
    - Benjamini-Yekutieli procedure, assuming arbitrary dependence between
        hypotheses (`arbitrary_dependence=True`).

    The function adds the following attributes to the results:

    - `pvalue_adj`: The adjusted p-value, which should be compared with
        the unadjusted FDR (`alpha`).
    - `alpha_adj`: The adjusted FDR, which should be compared with the unadjusted
        p-value (`pvalue`).
    - `null_rejected`: A binary indicator (`0` or `1`) that shows whether
        the null hypothesis is rejected.

    Args:
        experiment_results: Experiment results.
        metrics: Metrics included in the comparison.
            If `None`, all metrics are included.
        alpha: Significance level. If `None`, the value from global settings is used.
        arbitrary_dependence: If `True`, arbitrary dependence between hypotheses
            is assumed and Benjamini-Yekutieli procedure is performed.
            If `False`, non-negative correlation between hypotheses is assumed
            and Benjamini-Hochberg procedure is performed.

    Returns:
        The experiments results with adjusted p-values and alphas.

    Parameter defaults:
        Default for parameter `alpha` can be changed using the `config_context`
        and `set_context` functions.
        See the [Global configuration](https://tea-tasting.e10v.me/api/config/)
        reference for details.

    References:
        - [Multiple comparisons problem](https://en.wikipedia.org/wiki/Multiple_comparisons_problem).
        - [False discovery rate](https://en.wikipedia.org/wiki/False_discovery_rate).

    Examples:
        ```pycon
        >>> import polars as pl
        >>> import tea_tasting as tt

        >>> data = pl.concat((
        ...     tt.make_users_data(
        ...         seed=42,
        ...         orders_uplift=0.10,
        ...         revenue_uplift=0.15,
        ...         return_type="polars",
        ...     ),
        ...     tt.make_users_data(
        ...         seed=21,
        ...         orders_uplift=0.15,
        ...         revenue_uplift=0.20,
        ...         return_type="polars",
        ...     )
        ...         .filter(pl.col("variant").eq(1))
        ...         .with_columns(variant=pl.lit(2, pl.Int64)),
        ... ))
        >>> data
        shape: (6_046, 5)
        ┌──────┬─────────┬──────────┬────────┬─────────┐
        │ user ┆ variant ┆ sessions ┆ orders ┆ revenue │
        │ ---  ┆ ---     ┆ ---      ┆ ---    ┆ ---     │
        │ i64  ┆ i64     ┆ i64      ┆ i64    ┆ f64     │
        ╞══════╪═════════╪══════════╪════════╪═════════╡
        │ 0    ┆ 1       ┆ 2        ┆ 1      ┆ 9.58    │
        │ 1    ┆ 0       ┆ 2        ┆ 1      ┆ 6.43    │
        │ 2    ┆ 1       ┆ 2        ┆ 1      ┆ 8.3     │
        │ 3    ┆ 1       ┆ 2        ┆ 1      ┆ 16.65   │
        │ 4    ┆ 0       ┆ 1        ┆ 1      ┆ 7.14    │
        │ …    ┆ …       ┆ …        ┆ …      ┆ …       │
        │ 3989 ┆ 2       ┆ 4        ┆ 4      ┆ 34.93   │
        │ 3991 ┆ 2       ┆ 1        ┆ 0      ┆ 0.0     │
        │ 3992 ┆ 2       ┆ 3        ┆ 3      ┆ 27.96   │
        │ 3994 ┆ 2       ┆ 2        ┆ 1      ┆ 17.22   │
        │ 3998 ┆ 2       ┆ 3        ┆ 0      ┆ 0.0     │
        └──────┴─────────┴──────────┴────────┴─────────┘

        >>> experiment = tt.Experiment(
        ...     sessions_per_user=tt.Mean("sessions"),
        ...     orders_per_session=tt.RatioOfMeans("orders", "sessions"),
        ...     orders_per_user=tt.Mean("orders"),
        ...     revenue_per_user=tt.Mean("revenue"),
        ... )

        >>> # Results without correction.
        >>> results = experiment.analyze(data, control=0, all_variants=True)
        >>> results
        variants             metric control treatment rel_effect_size rel_effect_size_ci  pvalue
          (0, 1)  sessions_per_user    2.00      1.98          -0.66%      [-3.7%, 2.5%]   0.674
          (0, 1) orders_per_session   0.266     0.289            8.8%      [-0.89%, 19%]  0.0762
          (0, 1)    orders_per_user   0.530     0.573            8.0%       [-2.0%, 19%]   0.118
          (0, 1)   revenue_per_user    5.24      5.99             14%        [2.1%, 28%]  0.0211
          (0, 2)  sessions_per_user    2.00      2.02           0.98%      [-2.1%, 4.1%]   0.532
          (0, 2) orders_per_session   0.266     0.295             11%        [1.2%, 22%]  0.0273
          (0, 2)    orders_per_user   0.530     0.594             12%        [1.7%, 23%]  0.0213
          (0, 2)   revenue_per_user    5.24      6.25             19%        [6.6%, 33%] 0.00218

        >>> # Success metrics.
        >>> metrics = {"orders_per_user", "revenue_per_user"}

        >>> # Benjamini-Hochberg procedure,
        >>> # assuming non-negative correlation between hypotheses.
        >>> adjusted_results_fdr = tt.adjust_fdr(results, metrics)
        >>> adjusted_results_fdr
        comparison           metric control treatment rel_effect_size  pvalue pvalue_adj
            (0, 1)  orders_per_user   0.530     0.573            8.0%   0.118      0.118
            (0, 1) revenue_per_user    5.24      5.99             14%  0.0211     0.0284
            (0, 2)  orders_per_user   0.530     0.594             12%  0.0213     0.0284
            (0, 2) revenue_per_user    5.24      6.25             19% 0.00218    0.00872

        >>> # The adjusted confidence level alpha.
        >>> adjusted_results_fdr.with_keys((
        ...     "comparison",
        ...     "metric",
        ...     "control",
        ...     "treatment",
        ...     "rel_effect_size",
        ...     "pvalue",
        ...     "alpha_adj",
        ... ))
        comparison           metric control treatment rel_effect_size  pvalue alpha_adj
            (0, 1)  orders_per_user   0.530     0.573            8.0%   0.118    0.0500
            (0, 1) revenue_per_user    5.24      5.99             14%  0.0211    0.0375
            (0, 2)  orders_per_user   0.530     0.594             12%  0.0213    0.0375
            (0, 2) revenue_per_user    5.24      6.25             19% 0.00218    0.0375

        >>> # Benjamini-Yekutieli procedure,
        >>> # assuming arbitrary dependence between hypotheses.
        >>> tt.adjust_fdr(results, metrics, arbitrary_dependence=True)
        comparison           metric control treatment rel_effect_size  pvalue pvalue_adj
            (0, 1)  orders_per_user   0.530     0.573            8.0%   0.118      0.245
            (0, 1) revenue_per_user    5.24      5.99             14%  0.0211     0.0592
            (0, 2)  orders_per_user   0.530     0.594             12%  0.0213     0.0592
            (0, 2) revenue_per_user    5.24      6.25             19% 0.00218     0.0182

        ```
    """  # noqa: E501
    alpha = (
        tea_tasting.utils.auto_check(alpha, "alpha")
        if alpha is not None
        else tea_tasting.config.get_config("alpha")
    )
    arbitrary_dependence = tea_tasting.utils.check_scalar(
        arbitrary_dependence, "arbitrary_dependence", typ=bool)

    # results and metric_results refer to the same dicts.
    results, metric_results = _copy_results(experiment_results, metrics)
    method = _Benjamini(
        alpha=alpha,  # type: ignore
        m=len(metric_results),
        arbitrary_dependence=arbitrary_dependence,
    )
    # In-place update.
    _hochberg_stepup(metric_results, method.adjust)

    return MultipleComparisonsResults(results)


def adjust_fwer(
    experiment_results: tea_tasting.experiment.ExperimentResult | Mapping[
        object, tea_tasting.experiment.ExperimentResult],
    metrics: str | set[str] | Sequence[str] | None = None,
    *,
    alpha: float | None = None,
    arbitrary_dependence: bool = False,
    method: Literal["bonferroni", "sidak"] = "sidak",
) -> MultipleComparisonsResults:
    """Adjust p-value and alpha to control the family-wise error rate (FWER).

    The number of hypotheses tested is the total number of metrics included in
    the comparison in all experiment results. For example, if there are
    3 experiments with 2 metrics in each, the number of hypotheses is 6.

    The function performs one of the following procedures, depending on parameters:

    - Hochberg's step-up procedure, assuming non-negative correlation between
        hypotheses (`arbitrary_dependence=False`).
    - Holm's step-down procedure, assuming arbitrary dependence between
        hypotheses (`arbitrary_dependence=True`).

    The function adds the following attributes to the results:

    - `pvalue_adj`: The adjusted p-value, which should be compared with
        the unadjusted FDR (`alpha`).
    - `alpha_adj`: The adjusted FWER, which should be compared with the unadjusted
        p-value (`pvalue`).
    - `null_rejected`: A binary indicator (`0` or `1`) that shows whether
        the null hypothesis is rejected.

    Args:
        experiment_results: Experiment results.
        metrics: Metrics included in the comparison.
            If `None`, all metrics are included.
        alpha: Significance level. If `None`, the value from global settings is used.
        arbitrary_dependence: If `True`, arbitrary dependence between hypotheses
            is assumed and Holm's step-down procedure is performed.
            If `False`, non-negative correlation between hypotheses is assumed
            and Hochberg's step-up procedure is performed.
        method: Correction method, Bonferroni (`"bonferroni"`) or Šidák (`"sidak"`).

    Returns:
        The experiments results with adjusted p-values and alphas.

    Parameter defaults:
        Default for parameter `alpha` can be changed using the `config_context`
        and `set_context` functions.
        See the [Global configuration](https://tea-tasting.e10v.me/api/config/)
        reference for details.

    References:
        - [Multiple comparisons problem](https://en.wikipedia.org/wiki/Multiple_comparisons_problem).
        - [Family-wise error rate](https://en.wikipedia.org/wiki/Family-wise_error_rate).
        - [Holm–Bonferroni method](https://en.wikipedia.org/wiki/Holm%E2%80%93Bonferroni_method).

    Examples:
        ```pycon
        >>> import polars as pl
        >>> import tea_tasting as tt

        >>> data = pl.concat((
        ...     tt.make_users_data(
        ...         seed=42,
        ...         orders_uplift=0.10,
        ...         revenue_uplift=0.15,
        ...         return_type="polars",
        ...     ),
        ...     tt.make_users_data(
        ...         seed=21,
        ...         orders_uplift=0.15,
        ...         revenue_uplift=0.20,
        ...         return_type="polars",
        ...     )
        ...         .filter(pl.col("variant").eq(1))
        ...         .with_columns(variant=pl.lit(2, pl.Int64)),
        ... ))
        >>> data
        shape: (6_046, 5)
        ┌──────┬─────────┬──────────┬────────┬─────────┐
        │ user ┆ variant ┆ sessions ┆ orders ┆ revenue │
        │ ---  ┆ ---     ┆ ---      ┆ ---    ┆ ---     │
        │ i64  ┆ i64     ┆ i64      ┆ i64    ┆ f64     │
        ╞══════╪═════════╪══════════╪════════╪═════════╡
        │ 0    ┆ 1       ┆ 2        ┆ 1      ┆ 9.58    │
        │ 1    ┆ 0       ┆ 2        ┆ 1      ┆ 6.43    │
        │ 2    ┆ 1       ┆ 2        ┆ 1      ┆ 8.3     │
        │ 3    ┆ 1       ┆ 2        ┆ 1      ┆ 16.65   │
        │ 4    ┆ 0       ┆ 1        ┆ 1      ┆ 7.14    │
        │ …    ┆ …       ┆ …        ┆ …      ┆ …       │
        │ 3989 ┆ 2       ┆ 4        ┆ 4      ┆ 34.93   │
        │ 3991 ┆ 2       ┆ 1        ┆ 0      ┆ 0.0     │
        │ 3992 ┆ 2       ┆ 3        ┆ 3      ┆ 27.96   │
        │ 3994 ┆ 2       ┆ 2        ┆ 1      ┆ 17.22   │
        │ 3998 ┆ 2       ┆ 3        ┆ 0      ┆ 0.0     │
        └──────┴─────────┴──────────┴────────┴─────────┘

        >>> experiment = tt.Experiment(
        ...     sessions_per_user=tt.Mean("sessions"),
        ...     orders_per_session=tt.RatioOfMeans("orders", "sessions"),
        ...     orders_per_user=tt.Mean("orders"),
        ...     revenue_per_user=tt.Mean("revenue"),
        ... )

        >>> # Results without correction.
        >>> results = experiment.analyze(data, control=0, all_variants=True)
        >>> results
        variants             metric control treatment rel_effect_size rel_effect_size_ci  pvalue
          (0, 1)  sessions_per_user    2.00      1.98          -0.66%      [-3.7%, 2.5%]   0.674
          (0, 1) orders_per_session   0.266     0.289            8.8%      [-0.89%, 19%]  0.0762
          (0, 1)    orders_per_user   0.530     0.573            8.0%       [-2.0%, 19%]   0.118
          (0, 1)   revenue_per_user    5.24      5.99             14%        [2.1%, 28%]  0.0211
          (0, 2)  sessions_per_user    2.00      2.02           0.98%      [-2.1%, 4.1%]   0.532
          (0, 2) orders_per_session   0.266     0.295             11%        [1.2%, 22%]  0.0273
          (0, 2)    orders_per_user   0.530     0.594             12%        [1.7%, 23%]  0.0213
          (0, 2)   revenue_per_user    5.24      6.25             19%        [6.6%, 33%] 0.00218

        >>> # Success metrics.
        >>> metrics = {"orders_per_user", "revenue_per_user"}

        >>> # Hochberg's step-up procedure with Šidák correction,
        >>> # assuming non-negative correlation between hypotheses.
        >>> adjusted_results_fwer = tt.adjust_fwer(results, metrics)
        >>> adjusted_results_fwer
        comparison           metric control treatment rel_effect_size  pvalue pvalue_adj
            (0, 1)  orders_per_user   0.530     0.573            8.0%   0.118      0.118
            (0, 1) revenue_per_user    5.24      5.99             14%  0.0211     0.0422
            (0, 2)  orders_per_user   0.530     0.594             12%  0.0213     0.0422
            (0, 2) revenue_per_user    5.24      6.25             19% 0.00218    0.00869

        >>> # The adjusted confidence level alpha.
        >>> adjusted_results_fwer.with_keys((
        ...     "comparison",
        ...     "metric",
        ...     "control",
        ...     "treatment",
        ...     "rel_effect_size",
        ...     "pvalue",
        ...     "alpha_adj",
        ... ))
        comparison           metric control treatment rel_effect_size  pvalue alpha_adj
            (0, 1)  orders_per_user   0.530     0.573            8.0%   0.118    0.0500
            (0, 1) revenue_per_user    5.24      5.99             14%  0.0211    0.0253
            (0, 2)  orders_per_user   0.530     0.594             12%  0.0213    0.0253
            (0, 2) revenue_per_user    5.24      6.25             19% 0.00218    0.0253

        >>> # Holm's step-down procedure with Bonferroni correction,
        >>> # assuming arbitrary dependence between hypotheses.
        >>> tt.adjust_fwer(
        ...     results,
        ...     metrics,
        ...     arbitrary_dependence=True,
        ...     method="bonferroni",
        ... )
        comparison           metric control treatment rel_effect_size  pvalue pvalue_adj
            (0, 1)  orders_per_user   0.530     0.573            8.0%   0.118      0.118
            (0, 1) revenue_per_user    5.24      5.99             14%  0.0211     0.0634
            (0, 2)  orders_per_user   0.530     0.594             12%  0.0213     0.0634
            (0, 2) revenue_per_user    5.24      6.25             19% 0.00218    0.00872

        ```
    """  # noqa: E501, RUF002
    alpha = (
        tea_tasting.utils.auto_check(alpha, "alpha")
        if alpha is not None
        else tea_tasting.config.get_config("alpha")
    )
    method = tea_tasting.utils.check_scalar(
        method, "method", typ=str, in_={"sidak", "bonferroni"})
    arbitrary_dependence = tea_tasting.utils.check_scalar(
        arbitrary_dependence, "arbitrary_dependence", typ=bool)

    # results and metric_results refer to the same dicts.
    results, metric_results = _copy_results(experiment_results, metrics)
    method_cls = _Sidak if method == "sidak" else _Bonferroni
    method_ = method_cls(alpha=alpha, m=len(metric_results))  # type: ignore
    procedure = _holm_stepdown if arbitrary_dependence else _hochberg_stepup
    # In-place update.
    procedure(metric_results, method_.adjust)

    return MultipleComparisonsResults(results)


def _copy_results(
    experiment_results: tea_tasting.experiment.ExperimentResult | Mapping[
        object, tea_tasting.experiment.ExperimentResult],
    metrics: str | set[str] | Sequence[str] | None = None,
) -> tuple[
    dict[object, tea_tasting.experiment.ExperimentResult],
    list[dict[str, object]],
]:
    if isinstance(experiment_results, tea_tasting.experiment.ExperimentResult):
        experiment_results = {NO_NAME_COMPARISON: experiment_results}

    if metrics is not None:
        if isinstance(metrics, str):
            metrics = {metrics}
        elif not isinstance(metrics, set):
            metrics = set(metrics)

    copy_of_experiment_results = {}
    copy_of_metric_results = []
    for comparison, experiment_result in experiment_results.items():
        result = {}
        for metric, metric_result in experiment_result.items():
            if metrics is None or metric in metrics:
                copy_of_metric_result = (
                    metric_result.copy()
                    if isinstance(metric_result, dict)
                    else metric_result._asdict()
                )
                result |= {metric: copy_of_metric_result}
                copy_of_metric_results.append(copy_of_metric_result)
        copy_of_experiment_results |= {
            comparison: tea_tasting.experiment.ExperimentResult(result)}

    return copy_of_experiment_results, copy_of_metric_results


def _hochberg_stepup(
    metric_results: Sequence[dict[str, object]],
    adjust: Callable[[float, int], tuple[float, float]],
) -> None:
    pvalue_adj_max = 1
    alpha_adj_min = 0
    m = len(metric_results)
    for i, metric_result in enumerate(
        sorted(metric_results, key=lambda d: -d["pvalue"]),  # type: ignore
    ):
        k = m - i
        pvalue: float = metric_result["pvalue"]  # type: ignore
        pvalue_adj, alpha_adj = adjust(pvalue, k)

        if alpha_adj_min == 0 and pvalue <= alpha_adj:
            alpha_adj_min = alpha_adj
        alpha_adj = max(alpha_adj, alpha_adj_min)
        pvalue_adj = pvalue_adj_max = min(pvalue_adj, pvalue_adj_max)

        metric_result.update(
            pvalue_adj=pvalue_adj,
            alpha_adj=alpha_adj,
            null_rejected=int(pvalue <= alpha_adj),
        )


def _holm_stepdown(
    metric_results: Sequence[dict[str, object]],
    adjust: Callable[[float, int], tuple[float, float]],
) -> None:
    pvalue_adj_min = 0
    alpha_adj_max = 1
    for k, metric_result in enumerate(
        sorted(metric_results, key=lambda d: d["pvalue"]),  # type: ignore
        start=1,
    ):
        pvalue: float = metric_result["pvalue"]
        pvalue_adj, alpha_adj = adjust(pvalue, k)

        if alpha_adj_max == 1 and pvalue > alpha_adj:
            alpha_adj_max = alpha_adj
        alpha_adj = min(alpha_adj, alpha_adj_max)
        pvalue_adj = pvalue_adj_min = max(pvalue_adj, pvalue_adj_min)

        metric_result.update(
            pvalue_adj=pvalue_adj,
            alpha_adj=alpha_adj,
            null_rejected=int(pvalue <= alpha_adj),
        )


class _Correction(abc.ABC):
    @abc.abstractmethod
    def adjust(self, pvalue: float, k: int) -> tuple[float, float]:
        ...


class _Benjamini(_Correction):
    def __init__(self,
        alpha: float,
        m: int,
        *,
        arbitrary_dependence: bool,
    ) -> None:
        self.alpha = alpha
        self.m_adj_ = (
            m * sum(1 / i for i in range(1, m + 1))
            if arbitrary_dependence
            else m
        )

    def adjust(self, pvalue: float, k: int) -> tuple[float, float]:
        coef = self.m_adj_ / k
        return min(pvalue * coef, 1), self.alpha / coef


class _Bonferroni(_Correction):
    def __init__(self, alpha: float, m: int) -> None:
        self.alpha = alpha
        self.m = m

    def adjust(self, pvalue: float, k: int) -> tuple[float, float]:
        coef = self.m - k + 1
        return min(pvalue * coef, 1), self.alpha / coef


class _Sidak(_Correction):
    def __init__(self, alpha: float, m: int) -> None:
        self.alpha = alpha
        self.m = m

    def adjust(self, pvalue: float, k: int) -> tuple[float, float]:
        coef = self.m - k + 1
        return 1 - (1 - pvalue)**coef, 1 - (1 - self.alpha)**(1 / coef)



from __future__ import annotations

from typing import TYPE_CHECKING
import unittest.mock

import pytest

import tea_tasting.aggr
import tea_tasting.config
import tea_tasting.datasets
import tea_tasting.metrics.base
import tea_tasting.metrics.mean


if TYPE_CHECKING:
    import pyarrow as pa


@pytest.fixture
def data_arrow() -> pa.Table:
    return tea_tasting.datasets.make_users_data(n_users=100, covariates=True, seed=42)

@pytest.fixture
def data_aggr(data_arrow: pa.Table) -> dict[object, tea_tasting.aggr.Aggregates]:
    cols = (
        "sessions", "orders", "revenue",
        "sessions_covariate", "orders_covariate", "revenue_covariate",
    )
    return tea_tasting.aggr.read_aggregates(
        data_arrow,
        group_col="variant",
        has_count=True,
        mean_cols=cols,
        var_cols=cols,
        cov_cols=tuple(
            (col0, col1)
            for col0 in cols
            for col1 in cols
            if col0 < col1
        ),
    )

@pytest.fixture
def power_data_arrow() -> pa.Table:
    return tea_tasting.datasets.make_users_data(
        n_users=100, covariates=True, seed=42,
        sessions_uplift=0, orders_uplift=0, revenue_uplift=0,
    )

@pytest.fixture
def power_data_aggr(power_data_arrow: pa.Table) -> tea_tasting.aggr.Aggregates:
    cols = (
        "sessions", "orders", "revenue",
        "sessions_covariate", "orders_covariate", "revenue_covariate",
    )
    return tea_tasting.aggr.read_aggregates(
        power_data_arrow,
        group_col=None,
        has_count=True,
        mean_cols=cols,
        var_cols=cols,
        cov_cols=tuple(
            (col0, col1)
            for col0 in cols
            for col1 in cols
            if col0 < col1
        ),
    )


def test_ratio_of_means_init_default():
    metric = tea_tasting.metrics.mean.RatioOfMeans("a")
    assert metric.numer == "a"
    assert metric.denom is None
    assert metric.numer_covariate is None
    assert metric.denom_covariate is None
    assert metric.alternative == tea_tasting.config.get_config("alternative")
    assert metric.confidence_level == tea_tasting.config.get_config("confidence_level")
    assert metric.equal_var == tea_tasting.config.get_config("equal_var")
    assert metric.use_t == tea_tasting.config.get_config("use_t")
    assert metric.alpha == tea_tasting.config.get_config("alpha")
    assert metric.ratio == tea_tasting.config.get_config("ratio")
    assert metric.power == tea_tasting.config.get_config("power")
    assert metric.effect_size is None
    assert metric.rel_effect_size is None
    assert metric.n_obs is None

def test_ratio_of_means_init_custom():
    metric = tea_tasting.metrics.mean.RatioOfMeans(
        numer="a",
        denom="b",
        numer_covariate="c",
        denom_covariate="d",
        alternative="greater",
        confidence_level=0.9,
        equal_var=True,
        use_t=False,
        alpha=0.1,
        ratio=0.5,
        power=0.75,
        rel_effect_size=0.08,
        n_obs=(5_000, 10_000),
    )
    assert metric.numer == "a"
    assert metric.denom == "b"
    assert metric.numer_covariate == "c"
    assert metric.denom_covariate == "d"
    assert metric.alternative == "greater"
    assert metric.confidence_level == 0.9
    assert metric.equal_var is True
    assert metric.use_t is False
    assert metric.alpha == 0.1
    assert metric.ratio == 0.5
    assert metric.power == 0.75
    assert metric.effect_size is None
    assert metric.rel_effect_size == 0.08
    assert metric.n_obs == (5_000, 10_000)
    metric = tea_tasting.metrics.mean.RatioOfMeans("a", effect_size=(1, 0.2))
    assert metric.effect_size == (1, 0.2)

def test_ratio_of_means_init_config():
    with tea_tasting.config.config_context(
        alternative="greater",
        confidence_level=0.9,
        equal_var=True,
        use_t=False,
        alpha=0.1,
        ratio=0.5,
        power=0.75,
        n_obs=(5_000, 10_000),
    ):
        metric = tea_tasting.metrics.mean.RatioOfMeans("a")
    assert metric.alternative == "greater"
    assert metric.confidence_level == 0.9
    assert metric.equal_var is True
    assert metric.use_t is False
    assert metric.alpha == 0.1
    assert metric.ratio == 0.5
    assert metric.power == 0.75
    assert metric.n_obs == (5_000, 10_000)


def test_ratio_of_means_aggr_cols():
    metric = tea_tasting.metrics.mean.RatioOfMeans(
        numer="a",
        denom="b",
        numer_covariate="c",
    )
    aggr_cols = metric.aggr_cols
    assert isinstance(aggr_cols, tea_tasting.metrics.base.AggrCols)
    assert aggr_cols.has_count is True
    assert aggr_cols.mean_cols == ("a", "b", "c")
    assert aggr_cols.var_cols == ("a", "b", "c")
    assert len(aggr_cols.cov_cols) == 3
    assert set(aggr_cols.cov_cols) == {("a", "b"), ("a", "c"), ("b", "c")}


def test_ratio_of_means_analyze_frame(data_arrow: pa.Table):
    metric = tea_tasting.metrics.mean.RatioOfMeans(
        numer="orders",
        denom="sessions",
    )
    result = metric.analyze(data_arrow, 0, 1, variant="variant")
    assert isinstance(result, tea_tasting.metrics.mean.MeanResult)

def test_ratio_of_means_analyze_basic(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
):
    metric = tea_tasting.metrics.mean.RatioOfMeans(numer="orders")
    result = metric.analyze(data_aggr, 0, 1)
    assert isinstance(result, tea_tasting.metrics.mean.MeanResult)
    assert result.control == pytest.approx(0.5660377358490566)
    assert result.treatment == pytest.approx(0.3829787234042553)
    assert result.effect_size == pytest.approx(-0.18305901244480127)
    assert result.effect_size_ci_lower == pytest.approx(-0.5004359360326984)
    assert result.effect_size_ci_upper == pytest.approx(0.13431791114309583)
    assert result.rel_effect_size == pytest.approx(-0.32340425531914896)
    assert result.rel_effect_size_ci_lower == pytest.approx(-0.6593351111187646)
    assert result.rel_effect_size_ci_upper == pytest.approx(0.34378920945924096)
    assert result.pvalue == pytest.approx(0.2551230546709908)
    assert result.statistic == pytest.approx(-1.1447678040118034)

def test_ratio_of_means_analyze_ratio_greater_equal_var(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
):
    metric = tea_tasting.metrics.mean.RatioOfMeans(
        numer="orders",
        denom="sessions",
        alternative="greater",
        equal_var=True,
    )
    result = metric.analyze(data_aggr, 0, 1)
    assert isinstance(result, tea_tasting.metrics.mean.MeanResult)
    assert result.control == pytest.approx(0.2857142857142857)
    assert result.treatment == pytest.approx(0.20224719101123595)
    assert result.effect_size == pytest.approx(-0.08346709470304975)
    assert result.effect_size_ci_lower == pytest.approx(-0.21187246704082818)
    assert result.effect_size_ci_upper == float("inf")
    assert result.rel_effect_size == pytest.approx(-0.2921348314606741)
    assert result.rel_effect_size_ci_lower == pytest.approx(-0.5852384654309937)
    assert result.rel_effect_size_ci_upper == float("inf")
    assert result.pvalue == pytest.approx(0.8584716347525132)
    assert result.statistic == pytest.approx(-1.0794048813446926)

def test_ratio_of_means_analyze_ratio_less_use_norm(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
):
    metric = tea_tasting.metrics.mean.RatioOfMeans(
        numer="orders",
        denom="sessions",
        numer_covariate="orders_covariate",
        denom_covariate="sessions_covariate",
        alternative="less",
        use_t=False,
    )
    result = metric.analyze(data_aggr, 0, 1)
    assert isinstance(result, tea_tasting.metrics.mean.MeanResult)
    assert result.control == pytest.approx(0.25572348175909004)
    assert result.treatment == pytest.approx(0.23549786496156158)
    assert result.effect_size == pytest.approx(-0.020225616797528462)
    assert result.effect_size_ci_lower == float("-inf")
    assert result.effect_size_ci_upper == pytest.approx(0.07287939184944564)
    assert result.rel_effect_size == pytest.approx(-0.07909174651619377)
    assert result.rel_effect_size_ci_lower == float("-inf")
    assert result.rel_effect_size_ci_upper == pytest.approx(0.34578466493619153)
    assert result.pvalue == pytest.approx(0.3604265417728255)
    assert result.statistic == pytest.approx(-0.3573188986307722)


def test_ratio_of_means_solve_power_frame(power_data_arrow: pa.Table):
    metric = tea_tasting.metrics.mean.RatioOfMeans(
        numer="orders",
        denom="sessions",
        rel_effect_size=0.1,
    )
    results = metric.solve_power(power_data_arrow, "power")
    assert isinstance(results, tea_tasting.metrics.base.MetricPowerResults)
    assert len(results) == 1
    result = results[0]
    assert result.power > 0
    assert result.power < 1
    assert result.rel_effect_size == 0.1

def test_ratio_of_means_solve_power_power(power_data_aggr: tea_tasting.aggr.Aggregates):
    metric = tea_tasting.metrics.mean.RatioOfMeans(
        numer="orders",
        alternative="greater",
        rel_effect_size=0.1,
        n_obs=10_000,
    )
    results = metric.solve_power(power_data_aggr, "power")
    assert isinstance(results, tea_tasting.metrics.base.MetricPowerResults)
    assert len(results) == 1
    result = results[0]
    assert result.power == pytest.approx(0.895653144822902)
    assert result.effect_size == pytest.approx(0.1 * power_data_aggr.mean("orders"))
    assert result.rel_effect_size == 0.1
    assert result.n_obs == 10_000

def test_ratio_of_means_solve_power_effect_size(
    power_data_aggr: tea_tasting.aggr.Aggregates,
):
    metric = tea_tasting.metrics.mean.RatioOfMeans(
        numer="orders",
        alternative="less",
        n_obs=10_000,
    )
    results = metric.solve_power(power_data_aggr, "effect_size")
    assert isinstance(results, tea_tasting.metrics.base.MetricPowerResults)
    assert len(results) == 1
    result = results[0]
    assert result.power == 0.8
    assert result.effect_size == pytest.approx(-0.04027000268908317)
    assert result.rel_effect_size == pytest.approx(-0.08568085678528335)
    assert result.n_obs == 10_000

def test_ratio_of_means_solve_power_rel_effect_size(
    power_data_aggr: tea_tasting.aggr.Aggregates,
):
    metric = tea_tasting.metrics.mean.RatioOfMeans(numer="orders", n_obs=10_000)
    results = metric.solve_power(power_data_aggr, "rel_effect_size")
    assert isinstance(results, tea_tasting.metrics.base.MetricPowerResults)
    assert len(results) == 1
    result = results[0]
    assert result.power == 0.8
    assert result.effect_size == pytest.approx(0.04537464375458489)
    assert result.rel_effect_size == pytest.approx(0.09654179522252104)
    assert result.n_obs == 10_000

def test_ratio_of_means_solve_power_n_obs(power_data_aggr: tea_tasting.aggr.Aggregates):
    metric = tea_tasting.metrics.mean.RatioOfMeans(numer="orders", effect_size=0.05)
    results = metric.solve_power(power_data_aggr, "n_obs")
    assert isinstance(results, tea_tasting.metrics.base.MetricPowerResults)
    assert len(results) == 1
    result = results[0]
    assert result.power == 0.8
    assert result.effect_size == 0.05
    assert result.rel_effect_size == pytest.approx(
        0.05 / power_data_aggr.mean("orders"))
    assert result.n_obs == 8236

def test_ratio_of_means_solve_power_multi(power_data_aggr: tea_tasting.aggr.Aggregates):
    metric = tea_tasting.metrics.mean.RatioOfMeans(
        numer="orders",
        rel_effect_size=(0.05, 0.1),
        n_obs=(5_000, 10_000),
    )
    results = metric.solve_power(power_data_aggr, "power")
    assert isinstance(results, tea_tasting.metrics.base.MetricPowerResults)
    assert len(results) == 4

def test_ratio_of_means_solve_power_raises_effect_size(
    power_data_aggr: tea_tasting.aggr.Aggregates,
):
    metric = tea_tasting.metrics.mean.RatioOfMeans(numer="orders")
    with pytest.raises(ValueError, match="One of them should be defined"):
        metric.solve_power(power_data_aggr, "power")
    with pytest.raises(ValueError, match="Only one of them should be defined"):
        tea_tasting.metrics.mean.RatioOfMeans(
            numer="orders",
            effect_size=0.05,
            rel_effect_size=0.1,
        )

def test_ratio_of_means_solve_power_raises_max_iter(
    power_data_aggr: tea_tasting.aggr.Aggregates,
):
    metric = tea_tasting.metrics.mean.RatioOfMeans(numer="orders", rel_effect_size=0.01)
    with (
        unittest.mock.patch("tea_tasting.metrics.mean.MAX_ITER", 1),
        pytest.raises(RuntimeError, match="Maximum number of iterations"),
    ):
        metric.solve_power(power_data_aggr, "n_obs")


def test_mean_analyze(data_aggr: dict[object, tea_tasting.aggr.Aggregates]):
    metric = tea_tasting.metrics.mean.Mean(
        "orders",
        covariate="orders_covariate",
        alternative="greater",
        confidence_level=0.9,
        equal_var=True,
        use_t=False,
    )
    ratio_metric = tea_tasting.metrics.mean.RatioOfMeans(
        "orders",
        numer_covariate="orders_covariate",
        alternative="greater",
        confidence_level=0.9,
        equal_var=True,
        use_t=False,
    )
    assert metric.analyze(data_aggr, 0, 1) == ratio_metric.analyze(data_aggr, 0, 1)


def test_mean_solve_power(power_data_aggr: tea_tasting.aggr.Aggregates):
    metric = tea_tasting.metrics.mean.Mean(
        "orders",
        covariate="orders_covariate",
        alternative="greater",
        confidence_level=0.9,
        equal_var=True,
        use_t=False,
        rel_effect_size=0.1,
    )
    ratio_metric = tea_tasting.metrics.mean.RatioOfMeans(
        "orders",
        numer_covariate="orders_covariate",
        alternative="greater",
        confidence_level=0.9,
        equal_var=True,
        use_t=False,
        rel_effect_size=0.1,
    )
    assert metric.solve_power(power_data_aggr, "power") ==  ratio_metric.solve_power(
        power_data_aggr, "power")



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import tea_tasting.config


if TYPE_CHECKING:
    from collections.abc import Iterator


@pytest.fixture
def reset_config() -> Iterator[None]:
    try:
        yield
    finally:
        tea_tasting.config._config_var.set(tea_tasting.config._DEFAULT_CONFIG.copy())


@pytest.mark.usefixtures("reset_config")
def test_get_config():
    config = tea_tasting.config.get_config()
    assert config == tea_tasting.config._config_var.get()
    config["equal_var"] = not config["equal_var"]
    assert config != tea_tasting.config._config_var.get()

    assert (
        tea_tasting.config.get_config("equal_var") ==
        tea_tasting.config._config_var.get()["equal_var"]
    )


@pytest.mark.usefixtures("reset_config")
def test_set_config():
    tea_tasting.config.set_config(equal_var=True)
    assert tea_tasting.config._config_var.get()["equal_var"] is True

    tea_tasting.config.set_config(equal_var=False)
    assert tea_tasting.config._config_var.get()["equal_var"] is False


@pytest.mark.usefixtures("reset_config")
def test_config_context():
    old_equal_var = tea_tasting.config._config_var.get()["equal_var"]

    with tea_tasting.config.config_context(equal_var=not old_equal_var):
        assert tea_tasting.config._config_var.get() is not old_equal_var

    assert tea_tasting.config._config_var.get()["equal_var"] is old_equal_var



# pyright: reportAttributeAccessIssue=false
from __future__ import annotations

from typing import TYPE_CHECKING

import ibis
import numpy as np
import polars as pl
import pyarrow as pa
import pyarrow.compute as pc
import pytest

import tea_tasting.aggr
import tea_tasting.datasets


if TYPE_CHECKING:
    import ibis.expr.types  # noqa: TC004
    import pandas as pd


    Frame = ibis.expr.types.Table | pa.Table | pd.DataFrame | pl.LazyFrame


COUNT = 100
MEAN = {"x": 5.0, "y": 4}
VAR = {"x": 3.0, "y": 2}
COV = {("x", "y"): 1.0}

@pytest.fixture
def aggr() -> tea_tasting.aggr.Aggregates:
    return tea_tasting.aggr.Aggregates(
        count_=COUNT,
        mean_=MEAN,
        var_=VAR,
        cov_=COV,  # type: ignore
    )


@pytest.fixture
def data_arrow() -> pa.Table:
    return tea_tasting.datasets.make_users_data(n_users=100, seed=42)

@pytest.fixture
def data_pandas(data_arrow: pa.Table) -> pd.DataFrame:
    return data_arrow.to_pandas()

@pytest.fixture
def data_polars(data_arrow: pa.Table) -> pl.DataFrame:
    return pl.from_arrow(data_arrow)  # type: ignore

@pytest.fixture
def data_polars_lazy(data_polars: pl.DataFrame) -> pl.LazyFrame:
    return data_polars.lazy()

@pytest.fixture
def data_duckdb(data_arrow: pa.Table) -> ibis.expr.types.Table:
    return ibis.connect("duckdb://").create_table("data", data_arrow)

@pytest.fixture
def data_sqlite(data_arrow: pa.Table) -> ibis.expr.types.Table:
    return ibis.connect("sqlite://").create_table("data", data_arrow)

@pytest.fixture(params=[
    "data_arrow", "data_pandas",
    "data_polars", "data_polars_lazy",
    "data_duckdb", "data_sqlite",
])
def data(request: pytest.FixtureRequest) -> Frame:
    return request.getfixturevalue(request.param)


@pytest.fixture
def correct_aggr(data_arrow: pa.Table) -> tea_tasting.aggr.Aggregates:
    return tea_tasting.aggr.Aggregates(
        count_=data_arrow.num_rows,
        mean_={
            "sessions": pc.mean(data_arrow["sessions"]).as_py(),
            "orders": pc.mean(data_arrow["orders"]).as_py(),
        },
        var_={
            "sessions": pc.variance(data_arrow["sessions"], ddof=1).as_py(),
            "orders": pc.variance(data_arrow["orders"], ddof=1).as_py(),
        },
        cov_={
            ("orders", "sessions"): np.cov(
                data_arrow["sessions"].combine_chunks().to_numpy(zero_copy_only=False),
                data_arrow["orders"].combine_chunks().to_numpy(zero_copy_only=False),
                ddof=1,
            )[0, 1],
        },
    )

@pytest.fixture
def correct_aggrs(data_arrow: pa.Table) -> dict[int, tea_tasting.aggr.Aggregates]:
    variant_col = data_arrow["variant"]
    aggrs = {}
    for var in variant_col.unique().to_pylist():
        var_data = data_arrow.filter(pc.equal(variant_col, pa.scalar(var)))
        aggrs |= {var: tea_tasting.aggr.Aggregates(
            count_=var_data.num_rows,
            mean_={
                "sessions": pc.mean(var_data["sessions"]).as_py(),
                "orders": pc.mean(var_data["orders"]).as_py(),
            },
            var_={
                "sessions": pc.variance(var_data["sessions"], ddof=1).as_py(),
                "orders": pc.variance(var_data["orders"], ddof=1).as_py(),
            },
            cov_={
                ("orders", "sessions"): np.cov(
                    var_data["sessions"].combine_chunks().to_numpy(zero_copy_only=False),
                    var_data["orders"].combine_chunks().to_numpy(zero_copy_only=False),
                    ddof=1,
                )[0, 1],
            },
        )}
    return aggrs


def test_aggregates_init(aggr: tea_tasting.aggr.Aggregates):
    assert aggr.count_ == COUNT
    assert aggr.mean_ == MEAN
    assert aggr.var_ == VAR
    assert aggr.cov_ == COV

def test_aggregates_calls(aggr: tea_tasting.aggr.Aggregates):
    assert aggr.count() == COUNT
    assert aggr.mean("x") == MEAN["x"]
    assert aggr.mean("y") == MEAN["y"]
    assert aggr.var("x") == VAR["x"]
    assert aggr.mean("y") == MEAN["y"]
    assert aggr.cov("x", "y") == COV["x", "y"]

def test_aggregates_count_raises():
    aggr = tea_tasting.aggr.Aggregates(count_=None, mean_={}, var_={}, cov_={})
    with pytest.raises(RuntimeError):
        aggr.count()

def test_aggregates_none(aggr: tea_tasting.aggr.Aggregates):
    assert aggr.mean(None) == 1
    assert aggr.var(None) == 0
    assert aggr.cov(None, "y") == 0
    assert aggr.cov("x", None) == 0

def test_aggregates_ratio_var(aggr: tea_tasting.aggr.Aggregates):
    assert aggr.ratio_var("x", "y") == pytest.approx(0.2265625)

def test_aggregates_ratio_cov():
    aggr = tea_tasting.aggr.Aggregates(
        count_=None,
        mean_={"a": 8, "b": 7, "c": 6, "d": 5},
        var_={},
        cov_={("a", "c"): 4, ("a", "d"): 3, ("b", "c"): 2, ("b", "d"): 1},
    )
    assert aggr.ratio_cov("a", "b", "c", "d") == pytest.approx(-0.0146938775510204)

def test_aggregates_add(
    correct_aggr: tea_tasting.aggr.Aggregates,
    correct_aggrs: dict[int, tea_tasting.aggr.Aggregates],
):
    aggrs_add = correct_aggrs[0] + correct_aggrs[1]
    assert aggrs_add.count_ == pytest.approx(correct_aggr.count_)
    assert aggrs_add.mean_ == pytest.approx(correct_aggr.mean_)
    assert aggrs_add.var_ == pytest.approx(correct_aggr.var_)
    assert aggrs_add.cov_ == pytest.approx(correct_aggr.cov_)


def test_read_aggregates_groups(
    data: Frame,
    correct_aggrs: dict[int, tea_tasting.aggr.Aggregates],
):
    aggrs = tea_tasting.aggr.read_aggregates(
        data,
        group_col="variant",
        has_count=True,
        mean_cols=("sessions", "orders"),
        var_cols=("sessions", "orders"),
        cov_cols=(("sessions", "orders"),),
    )
    for i in (0, 1):
        assert aggrs[i].count_ == pytest.approx(correct_aggrs[i].count_)
        assert aggrs[i].mean_ == pytest.approx(correct_aggrs[i].mean_)
        assert aggrs[i].var_ == pytest.approx(correct_aggrs[i].var_)
        assert aggrs[i].cov_ == pytest.approx(correct_aggrs[i].cov_)

def test_read_aggregates_no_groups(
    data: Frame,
    correct_aggr: tea_tasting.aggr.Aggregates,
):
    aggr = tea_tasting.aggr.read_aggregates(
        data,
        group_col=None,
        has_count=True,
        mean_cols=("sessions", "orders"),
        var_cols=("sessions", "orders"),
        cov_cols=(("sessions", "orders"),),
    )
    assert aggr.count_ == pytest.approx(correct_aggr.count_)
    assert aggr.mean_ == pytest.approx(correct_aggr.mean_)
    assert aggr.var_ == pytest.approx(correct_aggr.var_)
    assert aggr.cov_ == pytest.approx(correct_aggr.cov_)

def test_read_aggregates_no_count(data_arrow: pa.Table):
    aggr = tea_tasting.aggr.read_aggregates(
        data_arrow,
        group_col=None,
        has_count=False,
        mean_cols=("sessions", "orders"),
        var_cols=(),
        cov_cols=(),
    )
    assert aggr.count_ is None
    assert aggr.var_ == {}
    assert aggr.cov_ == {}



from __future__ import annotations

from typing import NamedTuple

import pytest

import tea_tasting.experiment
import tea_tasting.multiplicity


class _MetricResult(NamedTuple):
    pvalue: float


@pytest.fixture
def experiment_results() -> dict[object, tea_tasting.experiment.ExperimentResult]:
    return {
        (0, 1): tea_tasting.experiment.ExperimentResult({
            "metric1": _MetricResult(pvalue=0.005),
            "metric2": {"pvalue": 0.06},
        }),
        (0, 2): tea_tasting.experiment.ExperimentResult({
            "metric2": _MetricResult(pvalue=0.007),
            "metric3": {"pvalue": 0.04},
        }),
        (0, 3): tea_tasting.experiment.ExperimentResult({
            "metric1": _MetricResult(pvalue=0.01),
            "metric3": {"pvalue": 0.03},
        }),
    }


def test_multiple_comparisons_results_to_dicts():
    results = tea_tasting.multiplicity.MultipleComparisonsResults({
        (0, 1): tea_tasting.experiment.ExperimentResult({
            "x": {"control": 10, "treatment": 11},
            "y": {"control": 20, "treatment": 22},
        }),
        "a/b": tea_tasting.experiment.ExperimentResult({
            "y": {"control": 30, "treatment": 33},
            "z": {"control": 40, "treatment": 44},
        }),
    })
    assert results.to_dicts() == (
        {"comparison": "(0, 1)", "metric": "x", "control": 10, "treatment": 11},
        {"comparison": "(0, 1)", "metric": "y", "control": 20, "treatment": 22},
        {"comparison": "a/b", "metric": "y", "control": 30, "treatment": 33},
        {"comparison": "a/b", "metric": "z", "control": 40, "treatment": 44},
    )


def test_adjust_fdr_default(
    experiment_results: dict[object, tea_tasting.experiment.ExperimentResult],
):
    results = tea_tasting.multiplicity.adjust_fdr(experiment_results)
    assert results[(0, 1)]["metric1"]["pvalue_adj"] == pytest.approx(0.02)  # type: ignore
    assert results[(0, 1)]["metric2"]["pvalue_adj"] == pytest.approx(0.06)  # type: ignore
    assert results[(0, 2)]["metric2"]["pvalue_adj"] == pytest.approx(0.02)  # type: ignore
    assert results[(0, 2)]["metric3"]["pvalue_adj"] == pytest.approx(0.048)  # type: ignore
    assert results[(0, 3)]["metric1"]["pvalue_adj"] == pytest.approx(0.02)  # type: ignore
    assert results[(0, 3)]["metric3"]["pvalue_adj"] == pytest.approx(0.045)  # type: ignore
    assert results[(0, 1)]["metric1"]["alpha_adj"] == pytest.approx(0.0416666666666667)  # type: ignore
    assert results[(0, 1)]["metric2"]["alpha_adj"] == pytest.approx(0.05)  # type: ignore
    assert results[(0, 2)]["metric2"]["alpha_adj"] == pytest.approx(0.0416666666666667)  # type: ignore
    assert results[(0, 2)]["metric3"]["alpha_adj"] == pytest.approx(0.0416666666666667)  # type: ignore
    assert results[(0, 3)]["metric1"]["alpha_adj"] == pytest.approx(0.0416666666666667)  # type: ignore
    assert results[(0, 3)]["metric3"]["alpha_adj"] == pytest.approx(0.0416666666666667)  # type: ignore
    assert results[(0, 1)]["metric1"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 1)]["metric2"]["null_rejected"] == 0  # type: ignore
    assert results[(0, 2)]["metric2"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 2)]["metric3"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 3)]["metric1"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 3)]["metric3"]["null_rejected"] == 1  # type: ignore

def test_adjust_fdr_arbitrary_dependence(
    experiment_results: dict[object, tea_tasting.experiment.ExperimentResult],
):
    results = tea_tasting.multiplicity.adjust_fdr(
        experiment_results,
        ("metric1", "metric2", "metric3"),
        arbitrary_dependence=True,
    )
    assert results[(0, 1)]["metric1"]["pvalue_adj"] == pytest.approx(0.049)  # type: ignore
    assert results[(0, 1)]["metric2"]["pvalue_adj"] == pytest.approx(0.147)  # type: ignore
    assert results[(0, 2)]["metric2"]["pvalue_adj"] == pytest.approx(0.049)  # type: ignore
    assert results[(0, 2)]["metric3"]["pvalue_adj"] == pytest.approx(0.1176)  # type: ignore
    assert results[(0, 3)]["metric1"]["pvalue_adj"] == pytest.approx(0.049)  # type: ignore
    assert results[(0, 3)]["metric3"]["pvalue_adj"] == pytest.approx(0.11025)  # type: ignore
    assert results[(0, 1)]["metric1"]["alpha_adj"] == pytest.approx(0.0102040816326531)  # type: ignore
    assert results[(0, 1)]["metric2"]["alpha_adj"] == pytest.approx(0.0204081632653061)  # type: ignore
    assert results[(0, 2)]["metric2"]["alpha_adj"] == pytest.approx(0.0102040816326531)  # type: ignore
    assert results[(0, 2)]["metric3"]["alpha_adj"] == pytest.approx(0.0170068027210884)  # type: ignore
    assert results[(0, 3)]["metric1"]["alpha_adj"] == pytest.approx(0.0102040816326531)  # type: ignore
    assert results[(0, 3)]["metric3"]["alpha_adj"] == pytest.approx(0.0136054421768707)  # type: ignore
    assert results[(0, 1)]["metric1"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 1)]["metric2"]["null_rejected"] == 0  # type: ignore
    assert results[(0, 2)]["metric2"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 2)]["metric3"]["null_rejected"] == 0  # type: ignore
    assert results[(0, 3)]["metric1"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 3)]["metric3"]["null_rejected"] == 0  # type: ignore

def test_adjust_fdr_single_experiment():
    results = tea_tasting.multiplicity.adjust_fdr(
        tea_tasting.experiment.ExperimentResult({
            "metric1": _MetricResult(pvalue=0.005),
            "metric2": {"pvalue": 0.06},
        }),
        arbitrary_dependence=True,
    )
    assert results["-"]["metric1"]["pvalue_adj"] == pytest.approx(0.015)  # type: ignore
    assert results["-"]["metric2"]["pvalue_adj"] == pytest.approx(0.09)  # type: ignore
    assert results["-"]["metric1"]["alpha_adj"] == pytest.approx(0.016666666666666666)  # type: ignore
    assert results["-"]["metric2"]["alpha_adj"] == pytest.approx(0.03333333333333333)  # type: ignore
    assert results["-"]["metric1"]["null_rejected"] == 1  # type: ignore
    assert results["-"]["metric2"]["null_rejected"] == 0  # type: ignore

def test_adjust_fdr_single_metric(
    experiment_results: dict[object, tea_tasting.experiment.ExperimentResult],
):
    results = tea_tasting.multiplicity.adjust_fdr(
        experiment_results,
        "metric1",
        arbitrary_dependence=True,
    )
    assert results[(0, 1)]["metric1"]["pvalue_adj"] == pytest.approx(0.015)  # type: ignore
    assert results[(0, 3)]["metric1"]["pvalue_adj"] == pytest.approx(0.015)  # type: ignore
    assert results[(0, 1)]["metric1"]["alpha_adj"] == pytest.approx(0.03333333333333333)  # type: ignore
    assert results[(0, 3)]["metric1"]["alpha_adj"] == pytest.approx(0.03333333333333333)  # type: ignore
    assert results[(0, 1)]["metric1"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 3)]["metric1"]["null_rejected"] == 1  # type: ignore


def test_adjust_fwer_default(
    experiment_results: dict[object, tea_tasting.experiment.ExperimentResult],
):
    results = tea_tasting.multiplicity.adjust_fwer(experiment_results)
    assert results[(0, 1)]["metric1"]["pvalue_adj"] == pytest.approx(0.0296274906437343)  # type: ignore
    assert results[(0, 1)]["metric2"]["pvalue_adj"] == pytest.approx(0.0600000000000001)  # type: ignore
    assert results[(0, 2)]["metric2"]["pvalue_adj"] == pytest.approx(0.0345134180118071)  # type: ignore
    assert results[(0, 2)]["metric3"]["pvalue_adj"] == pytest.approx(0.0600000000000001)  # type: ignore
    assert results[(0, 3)]["metric1"]["pvalue_adj"] == pytest.approx(0.0394039900000001)  # type: ignore
    assert results[(0, 3)]["metric3"]["pvalue_adj"] == pytest.approx(0.0600000000000001)  # type: ignore
    assert results[(0, 1)]["metric1"]["alpha_adj"] == pytest.approx(0.0127414550985662)  # type: ignore
    assert results[(0, 1)]["metric2"]["alpha_adj"] == pytest.approx(0.05)  # type: ignore
    assert results[(0, 2)]["metric2"]["alpha_adj"] == pytest.approx(0.0127414550985662)  # type: ignore
    assert results[(0, 2)]["metric3"]["alpha_adj"] == pytest.approx(0.0253205655191037)  # type: ignore
    assert results[(0, 3)]["metric1"]["alpha_adj"] == pytest.approx(0.0127414550985662)  # type: ignore
    assert results[(0, 3)]["metric3"]["alpha_adj"] == pytest.approx(0.0169524275084415)  # type: ignore
    assert results[(0, 1)]["metric1"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 1)]["metric2"]["null_rejected"] == 0  # type: ignore
    assert results[(0, 2)]["metric2"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 2)]["metric3"]["null_rejected"] == 0  # type: ignore
    assert results[(0, 3)]["metric1"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 3)]["metric3"]["null_rejected"] == 0  # type: ignore

def test_adjust_fwer_arbitrary_dependence(
    experiment_results: dict[object, tea_tasting.experiment.ExperimentResult],
):
    results = tea_tasting.multiplicity.adjust_fwer(
        experiment_results,
        arbitrary_dependence=True,
    )
    assert results[(0, 1)]["metric1"]["pvalue_adj"] == pytest.approx(0.0296274906437343)  # type: ignore
    assert results[(0, 1)]["metric2"]["pvalue_adj"] == pytest.approx(0.087327)  # type: ignore
    assert results[(0, 2)]["metric2"]["pvalue_adj"] == pytest.approx(0.0345134180118071)  # type: ignore
    assert results[(0, 2)]["metric3"]["pvalue_adj"] == pytest.approx(0.087327)  # type: ignore
    assert results[(0, 3)]["metric1"]["pvalue_adj"] == pytest.approx(0.0394039900000001)  # type: ignore
    assert results[(0, 3)]["metric3"]["pvalue_adj"] == pytest.approx(0.087327)  # type: ignore
    assert results[(0, 1)]["metric1"]["alpha_adj"] == pytest.approx(0.0085124446108471)  # type: ignore
    assert results[(0, 1)]["metric2"]["alpha_adj"] == pytest.approx(0.0169524275084415)  # type: ignore
    assert results[(0, 2)]["metric2"]["alpha_adj"] == pytest.approx(0.0102062183130115)  # type: ignore
    assert results[(0, 2)]["metric3"]["alpha_adj"] == pytest.approx(0.0169524275084415)  # type: ignore
    assert results[(0, 3)]["metric1"]["alpha_adj"] == pytest.approx(0.0127414550985662)  # type: ignore
    assert results[(0, 3)]["metric3"]["alpha_adj"] == pytest.approx(0.0169524275084415)  # type: ignore
    assert results[(0, 1)]["metric1"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 1)]["metric2"]["null_rejected"] == 0  # type: ignore
    assert results[(0, 2)]["metric2"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 2)]["metric3"]["null_rejected"] == 0  # type: ignore
    assert results[(0, 3)]["metric1"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 3)]["metric3"]["null_rejected"] == 0  # type: ignore

def test_adjust_fwer_bonferroni(
    experiment_results: dict[object, tea_tasting.experiment.ExperimentResult],
):
    results = tea_tasting.multiplicity.adjust_fwer(
        experiment_results,
        method="bonferroni",
    )
    assert results[(0, 1)]["metric1"]["pvalue_adj"] == pytest.approx(0.03)  # type: ignore
    assert results[(0, 1)]["metric2"]["pvalue_adj"] == pytest.approx(0.06)  # type: ignore
    assert results[(0, 2)]["metric2"]["pvalue_adj"] == pytest.approx(0.035)  # type: ignore
    assert results[(0, 2)]["metric3"]["pvalue_adj"] == pytest.approx(0.06)  # type: ignore
    assert results[(0, 3)]["metric1"]["pvalue_adj"] == pytest.approx(0.04)  # type: ignore
    assert results[(0, 3)]["metric3"]["pvalue_adj"] == pytest.approx(0.06)  # type: ignore
    assert results[(0, 1)]["metric1"]["alpha_adj"] == pytest.approx(0.0125)  # type: ignore
    assert results[(0, 1)]["metric2"]["alpha_adj"] == pytest.approx(0.05)  # type: ignore
    assert results[(0, 2)]["metric2"]["alpha_adj"] == pytest.approx(0.0125)  # type: ignore
    assert results[(0, 2)]["metric3"]["alpha_adj"] == pytest.approx(0.025)  # type: ignore
    assert results[(0, 3)]["metric1"]["alpha_adj"] == pytest.approx(0.0125)  # type: ignore
    assert results[(0, 3)]["metric3"]["alpha_adj"] == pytest.approx(0.0166666666666667)  # type: ignore
    assert results[(0, 1)]["metric1"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 1)]["metric2"]["null_rejected"] == 0  # type: ignore
    assert results[(0, 2)]["metric2"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 2)]["metric3"]["null_rejected"] == 0  # type: ignore
    assert results[(0, 3)]["metric1"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 3)]["metric3"]["null_rejected"] == 0  # type: ignore

def test_adjust_fwer_arbitrary_dependence_bonferroni(
    experiment_results: dict[object, tea_tasting.experiment.ExperimentResult],
):
    results = tea_tasting.multiplicity.adjust_fwer(
        experiment_results,
        arbitrary_dependence=True,
        method="bonferroni",
    )
    assert results[(0, 1)]["metric1"]["pvalue_adj"] == pytest.approx(0.03)  # type: ignore
    assert results[(0, 1)]["metric2"]["pvalue_adj"] == pytest.approx(0.09)  # type: ignore
    assert results[(0, 2)]["metric2"]["pvalue_adj"] == pytest.approx(0.035)  # type: ignore
    assert results[(0, 2)]["metric3"]["pvalue_adj"] == pytest.approx(0.09)  # type: ignore
    assert results[(0, 3)]["metric1"]["pvalue_adj"] == pytest.approx(0.04)  # type: ignore
    assert results[(0, 3)]["metric3"]["pvalue_adj"] == pytest.approx(0.09)  # type: ignore
    assert results[(0, 1)]["metric1"]["alpha_adj"] == pytest.approx(0.00833333333333333)  # type: ignore
    assert results[(0, 1)]["metric2"]["alpha_adj"] == pytest.approx(0.0166666666666667)  # type: ignore
    assert results[(0, 2)]["metric2"]["alpha_adj"] == pytest.approx(0.01)  # type: ignore
    assert results[(0, 2)]["metric3"]["alpha_adj"] == pytest.approx(0.0166666666666667)  # type: ignore
    assert results[(0, 3)]["metric1"]["alpha_adj"] == pytest.approx(0.0125)  # type: ignore
    assert results[(0, 3)]["metric3"]["alpha_adj"] == pytest.approx(0.0166666666666667)  # type: ignore
    assert results[(0, 1)]["metric1"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 1)]["metric2"]["null_rejected"] == 0  # type: ignore
    assert results[(0, 2)]["metric2"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 2)]["metric3"]["null_rejected"] == 0  # type: ignore
    assert results[(0, 3)]["metric1"]["null_rejected"] == 1  # type: ignore
    assert results[(0, 3)]["metric3"]["null_rejected"] == 0  # type: ignore
